/*
** Module   : ControlFlowOps.java
** Abstract : Progress compatible control flow helper methods
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -----------------------Description-----------------------------------
** 001 GES 20050616   @21506 Created initial version, supporting PAUSE.
** 002 SIY 20050815   @22152 Added invoke() method used by RUN VALUE().
** 003 GES 20050829   @22381 Added error condition processing.
** 004 GES 20060117   @23955 Removed PAUSE (it now resides in UI code).
** 005 GES 20060123   @24026 Added return-value support as context-local.
** 006 GES 20060314   @25063 Fixed exception processing in invoke().
** 007 ECF 20060329   @25259 Removed fillInStackTrace calls to exceptions
**                           caught and re-thrown in invoke(). These were
**                           masking the true origins of the exceptions.
** 008 GES 20060427   @25774 Safety code to allow wrapper classes to be
**                           passed as a program name.
** 009 GES 20061030   @30804 Better processing of condition exceptions.
** 010 GES 20070220   @32194 Added support for RUN IN.
** 011 GES 20080619   @38885 If a RUN VALUE (actually any RUN statement)
**                           cannot find the procedure to execute, a STOP
**                           condition is raised (after some error text
**                           has been displayed on the terminal), not an
**                           ERROR. This means that silent error mode does
**                           not apply, the exception is always thrown and
**                           the ERROR-STATUS handle is not updated.
** 012 GES 20080619   @39166 Fix for RUN VALUE of an internal procedure in
**                           compact mode.
** 013 SIY 20080804   @39280 Changed signature of simple invoke(). This
**                           enables correct calls to internal procedures
**                           without redundant instantiation of class and
**                           greatly simplifies compact mode.
**                           Some minor cleanups and potential NPE fix.
** 014 GES 20090422   @41906 Converted to standard string formatting.
** 015 GES 20101005          Added a stub (non-functional) for a RUN statement
**                           that references a PROCEDURE EXTERNAL definition.
**                           Added a stub for DELETE PROCEDURE.
** 016 CA  20130116          Major rework to provide support for IN handle and
**                           IN SUPER functions and procedures, plus RUN SUPER
**                           and SUPER statements.
** 017 CA  20130126          Added support for RELEASE EXTERNAL (converts to
**                           a ControlFlowOps.release(<library>) call, which
**                           is only a stub at this time).
** 018 CA  20130213          Fixed some cases of extent parameter validation.
** 019 CS  20130213          Added APIs supporting the ON server and ASYNC clauses for the RUN 
**                           statement.
** 020 CA  20130223          Added conversion support for the RUN ... SET hPort ON hWebServer stmt
**                           to invoke web services.
** 021 CA  20130315          Added support for cases when a WrappedResource is passed to a handle
**                           parameter.
** 022 OM  20130404          Fixed unknown parameter in function/procedure calls.
**                           Abort function/procedure call if errors occurred when converting 
**                           compatible parameters (ex: 32bit overflow from decimal/int64).
** 023 CA  20130529          Added support for RUN ... [PERSISTENT [handle]] ON SERVER server and
**                           RUN ... IN remoteProcHandle. 
** 024 CA  20130708          Fixed validation of arrays with non-BDT components (these will be 
**                           validated by attempting to instantiate a wrapper using a c'tor which
**                           accepts this array as a parameter).
** 025 CA  20130813          Added support for async procedure calls.
** 026 SVL 20131203          Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 027 CA  20131116          Added support for the FUNCTION ... MAP TO clause.  Fixed runtime for 
**                           RUN SUPER and SUPER() calls, in search-target and search-self modes.
** 028 EVK 20131204          Made enum ArgValidationErrors public. Added method
**                           reachableInternalEntry() to check reachability of procedure or
**                           function in 4gl handle.
** 029 CA  20131220          Refactored the async invocation code and the remote invocation code
**                           to support both web service and appserver requests.
** 030 GES 20140116          Added support for native library calls.
** 031 CA  20140218          Fixed SOURCE- and TARGET-PROCEDURE when invoking external programs.
**                           Error messages must use the program name as it was used at the RUN
**                           statement. Internally, the absolute name must be used (i.e. when 
**                           invoking SourceNameMapper APIs).
** 032 ECF 20140613          Replaced renamed RecordBuffer method in getSignature.
** 033 MAG 20140626          Prevent fault messages to be displayed twice when invoke a web service.
** 034 VMN 20140702          Added method convertValue() for correct processing unknown value
**                           in invoke... methods.
** 035 HC  20140613          Improved extent parameter support and added logging.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.*;
import com.goldencode.p2j.library.*;
import com.goldencode.p2j.persist.RecordBuffer;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.proxy.ProxyFactory;

/**
 * Progress compatible control-flow related static utility methods. Current
 * implementation contains support for the:
 * <p>
 * <pre>
 * Progress 4GL Feature                  Java Implementation
 * -----------------------------------   ------------------------------------
 * RUN [ &lt;procedure&gt; | VALUE() ]   invoke()
 * RUN ... IN &lt;handle&gt;             invokeIn()
 * RUN ... PERSIST                       invokePersistent()
 * RUN ... PERSIST SET &lt;handle&gt;    invokePersistentSet() 
 * RETURN-VALUE "global variable"        getReturnValue()/setReturnValue()
 * RUN SUPER                             runSuper()
 * simple function invocation            invokeFunction()
 * FUNCTION ... IN handle                invokeFunctionIn()
 * DYNAMIC-FUNCTION                      invokeDynamicFunction()
 * DYNAMIC-FUNCTION(... IN)              invokeDynamicFunctionIn()
 * SUPER()                               runSuper()
 * RELEASE EXTERNAL [PROCEDURE]          release()
 * </pre>
 */
public class ControlFlowOps
{
   /** Enum with the list of possible argument validation errors. */
   public static enum ArgValidationErrors 
   {
      NO_ERROR,
      INCORRECT_COUNT,
      UNEXPECTED_PARAM,
      INCORRECT_TYPES,
      DYN_ARRAY_FOR_NON_ARRAY,
      NON_ARRAY_FOR_DYN_ARRAY,
      ARRAY_FOR_NON_ARRAY,
      NON_ARRAY_FOR_ARRAY,
      DEFER_VALIDATION;

      /**
       * In case of array-related validation error, this represents the extent set at the 
       * procedure/function parameter.
       */
      int expectedExtent = 0;
      
      /**
       * In case of array-related validation error, this represents the extent set at the 
       * variable/field passed to the function/procedure call.
       */
      int paramExtent = 0;
   };

   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer();
   
   /**
    * Attempts to unlpad the specified library from memory (the RELEASE EXTERNAL statement).
    * This code just delegates the processing to the {@link NativeInvoker#release} method.
    * 
    * @param    libname
    *           The name of the library to attempt to unload.
    */
   public static void release(character libname)
   {
      if (libname == null)
      {
         // quick/silent out, there is no way to unload a library in this case
         return;
      }
      
      if (libname.isUnknown())
      {
         // strange case where the error is raised in normal mode (without any message
         // display) and in silent mode only the error flag is set (but normal details
         // are not recorded)
         ErrorManager.noRecordOrThrowError("Unknown cannot be given for a libname!");
      }
      else
      {
         release(libname.toStringMessage());
      }
   }
   
   /**
    * Attempts to unload the specified library from memory (the RELEASE EXTERNAL statement).
    * This code just delegates the processing to the {@link NativeInvoker#release} method.
    * 
    * @param    libname
    *           The name of the library to attempt to unload. If <code>null</code> or the
    *           empty string are passed, this is a no-operation.
    */
   public static void release(String libname)
   {
      if (libname == null || libname.length() == 0)
      {
         // quick/silent out, there is no way to unload a library in this case
         return;
      }
      
      NativeInvoker.release(libname);
   }
   
   /**
    * Gets the most recent RETURN-VALUE for the current user's session.
    *
    * @return   The RETURN-VALUE which defaults to the empty string if a
    *           the equivalent of a Progress 4GL RUN statement has executed
    *           but no explicit RETURN-VALUE was set.
    */
   public static character getReturnValue()
   {
      return work.obtain().returnValue;
   }
   
   /**
    * Sets the most recent RETURN-VALUE for the current user's session.
    *
    * @param    returnValue
    *           The RETURN-VALUE to set or <code>null</code> to assign the
    *           empty string.
    */
   public static void setReturnValue(character returnValue)
   {
      if (returnValue == null)
      {
         returnValue = new character("");
      }
      
      work.obtain().returnValue.assign(returnValue);
   }
   
   /**
    * Sets the most recent RETURN-VALUE for the current user's session.
    *
    * @param    returnValue
    *           The RETURN-VALUE to set or <code>null</code> to assign the
    *           empty string.
    */
   public static void setReturnValue(String returnValue)
   {
      setReturnValue(new character(returnValue == null ? "" : returnValue));
   }

   /**
    * Execute the RUN SUPER statement or SUPER function. Runtime decides on
    * whether this is a function or procedure, based on the 
    * {@link TransactionManager#nearestExecutingBlock() nearest block}.
    * 
    * @param    iename
    *           The internal-entry's legacy 4GL name.
    * @param    args
    *           The argument list.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default, on error or for
    *           procedures.
    */
   public static BaseDataType runSuper(String iename, Object... args)
   {
      BlockDefinition block = TransactionManager.nearestExecutingBlock();
      boolean function = block.type == BlockType.FUNCTION;
      handle target = ProcedureManager.getSuperHandle(iename, function);
      if (target == null)
      {
         return new unknown();
      }

      character name = new character(iename);
      
      if (function)
      {
         return invokeFunctionImpl(name, target, false, true, null, args);
      }
      else
      {
         invokeInImpl(name, target, true, null, args);
         return new unknown();
      }
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeWithMode(String    name,
                                     String    modes, 
                                     Object... args) 
   {
      invokeWithMode(new character(name), modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeWithMode(character name,
                                     String    modes,
                                     Object... args)
   {
      invokeImpl((handle) null, null, name, false, false, false, false, modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invoke(String name, Object... args) 
   {
      invoke(new character(name), args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invoke(character name, Object... args)
   {
      invokeImpl((handle) null, null, name, false, false, false, false, null, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteWithMode(String    name, 
                                           handle    portHandle,
                                           handle    remoteHandle,
                                           boolean   transactionDistinct,
                                           String    modes, 
                                           Object... args) 
   {
      invokeRemoteWithMode(new character(name), portHandle, 
                           remoteHandle, transactionDistinct, 
                           modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteWithMode(character name, 
                                           handle    portHandle,
                                           handle    remoteHandle,
                                           boolean   transactionDistinct,
                                           String    modes, 
                                           Object... args) 
   {
      invokeRemoteWithModeImpl(null, name,
                               portHandle, remoteHandle,
                               transactionDistinct,
                               modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN statement).
    * 
    * @param    asyncReq
    *           When not-null, an {@link AsyncRequestImpl} instance for the associated async
    *           request.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   private static void invokeRemoteWithModeImpl(AsyncRequestImpl asyncReq,
                                                character        name, 
                                                handle           portHandle,
                                                handle           remoteHandle,
                                                boolean          transactionDistinct,
                                                String           modes, 
                                                Object...        args) 
   {
      if (portHandle != null)
      {
         // this is a RUN ... SET ... ON SERVER webService
         invokePersistentImpl(asyncReq, remoteHandle, name, null, portHandle,
                              transactionDistinct, modes, args);
      }
      else
      {
         // this is a RUN ... ON SERVER appServer
         invokeImpl(asyncReq, remoteHandle, null, name,
                    false, false, false,
                    transactionDistinct, modes, args);
      }
   }

   /**
    * Invoke the procedure with the given name on a remote server (emitted for a standard RUN ... 
    * ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemote(String    name, 
                                   handle    portHandle,
                                   handle    remoteHandle, 
                                   boolean   transactionDistinct, 
                                   Object... args)
   {
      invokeRemote(new character(name), portHandle, remoteHandle, transactionDistinct, args);
   }

   /**
    * Invoke the procedure with the given name on a remote server  (emitted for a standard RUN ... 
    * ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemote(character name,
                                   handle    portHandle,
                                   handle    remoteHandle, 
                                   boolean   transactionDistinct, 
                                   Object... args)
   {
      invokeRemoteWithMode(new character(name), portHandle, 
                           remoteHandle, transactionDistinct, 
                           null, args);
   }

   /**
    * Invoke the procedure with the given name on a remote server (emitted for a standard RUN ... 
    * ON ... ASYNC SET ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsyncWithMode(String    name, 
                                                handle    portHandle,
                                                handle    remoteHandle, 
                                                boolean   transactionDistinct,
                                                handle    asyncHandle,
                                                String    eventProcName,
                                                handle    inEventprocHandle,
                                                String    modes, 
                                                Object... args)
   {
      invokeRemoteAsyncWithMode(new character(name), portHandle, 
                                remoteHandle, transactionDistinct, 
                                asyncHandle, new character(eventProcName), inEventprocHandle, 
                                modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsyncWithMode(character name, 
                                                handle    portHandle,
                                                handle    remoteHandle, 
                                                boolean   transactionDistinct,
                                                handle    asyncHandle,
                                                String    eventProcName,
                                                handle    inEventprocHandle,
                                                String    modes, 
                                                Object... args)
   {
      invokeRemoteAsyncWithMode(name, portHandle, remoteHandle, transactionDistinct,
                                asyncHandle, new character(eventProcName), inEventprocHandle, 
                                modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsyncWithMode(String    name, 
                                                handle    portHandle,
                                                handle    remoteHandle, 
                                                boolean   transactionDistinct,
                                                handle    asyncHandle,
                                                character eventProcName,
                                                handle    inEventprocHandle,
                                                String    modes, 
                                                Object... args)
   {
      invokeRemoteAsyncWithMode(new character(name), portHandle, 
                                remoteHandle, transactionDistinct, 
                                asyncHandle, eventProcName, inEventprocHandle,
                                modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsyncWithMode(character name, 
                                                handle    portHandle,
                                                handle    remoteHandle, 
                                                boolean   transactionDistinct,
                                                handle    asyncHandle,
                                                character eventProcName,
                                                handle    inEventprocHandle,
                                                String    modes, 
                                                Object... args)
   {
      AsyncCall remoteRequest = new AsyncCall(remoteHandle, portHandle, name, new handle(),
                                              asyncHandle, eventProcName, inEventprocHandle,
                                              transactionDistinct,
                                              modes, args)
      {
         @Override
         public void run()
         {
            // use a copy here, to not allow the user to change the reference
            invokeRemoteWithModeImpl(asyncReq, pName,
                                     portHandle, hServer,
                                     transactionDistinct,
                                     asyncReq.getModes(), asyncReq.getArgs());
         }
      };

      invokeAsyncImpl(remoteRequest, remoteHandle, asyncHandle, new handle());
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsync(String    name, 
                                        handle    portHandle,
                                        handle    remoteHandle, 
                                        boolean   transactionDistinct,
                                        handle    asyncHandle,
                                        String    eventProcName,
                                        handle    inEventprocHandle,
                                        Object... args)
   {
      invokeRemoteAsync(new character(name), portHandle, remoteHandle , transactionDistinct,
                        asyncHandle, new character(eventProcName), inEventprocHandle,
                        args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsync(character name, 
                                        handle    portHandle,
                                        handle    remoteHandle, 
                                        boolean   transactionDistinct,
                                        handle    asyncHandle,
                                        String    eventProcName,
                                        handle    inEventprocHandle,
                                        Object... args)
   {
      invokeRemoteAsync(name, portHandle, remoteHandle, transactionDistinct,
                        asyncHandle, new character(eventProcName), inEventprocHandle,
                        args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsync(String    name, 
                                        handle    portHandle,
                                        handle    remoteHandle, 
                                        boolean   transactionDistinct,
                                        handle    asyncHandle,
                                        character eventProcName,
                                        handle    inEventprocHandle,
                                        Object... args)
   {
      invokeRemoteAsync(new character(name), portHandle, remoteHandle , transactionDistinct,
                        asyncHandle, eventProcName, inEventprocHandle,
                        args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsync(character name, 
                                        handle    portHandle,
                                        handle    remoteHandle, 
                                        boolean   transactionDistinct,
                                        handle    asyncHandle,
                                        character eventProcName,
                                        handle    inEventprocHandle,
                                        Object... args)
   {
      invokeRemoteAsyncWithMode(name, portHandle, remoteHandle, transactionDistinct,
                                asyncHandle, eventProcName, inEventprocHandle,
                                null, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInWithMode(String    name,
                                       handle    h,
                                       String    modes,
                                       Object... args)
   {
      invokeInWithMode(new character(name), h, modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInWithMode(character name,
                                       handle    h,
                                       String    modes,
                                       Object... args)
   {
      invokeInImpl(name, h, false, modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeIn(String name, handle h, Object... args)
   {
      invokeIn(new character(name), h, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeIn(character name, handle h, Object... args)
   {
      invokeInImpl(name, h, false, null, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsyncWithMode(String    name,
                                            handle    h,
                                            handle    asyncHandle, 
                                            String    eventProcName,
                                            handle    inEventprocHandle,
                                            String    modes,
                                            Object... args)
   {
      invokeInAsyncWithMode(new character(name), h,
                            asyncHandle, new  character(eventProcName), inEventprocHandle,
                            modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN ASYNC.. handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsyncWithMode(character name,
                                            handle    h,
                                            handle    asyncHandle, 
                                            String    eventProcName,
                                            handle    inEventprocHandle,
                                            String    modes,
                                            Object... args)
   {
      invokeInAsyncWithMode(name, h,
                            asyncHandle, new character(eventProcName), inEventprocHandle,
                            modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN ASYNC.. handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsyncWithMode(String    name,
                                            handle    h,
                                            handle    asyncHandle, 
                                            character eventProcName,
                                            handle    inEventprocHandle,
                                            String    modes,
                                            Object... args)
   {
      invokeInAsyncWithMode(new character(name), h, asyncHandle, eventProcName, inEventprocHandle,
                            modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsyncWithMode(character name,
                                            handle    h,
                                            handle    asyncHandle, 
                                            character eventProcName,
                                            handle    inEventprocHandle,
                                            String    modes,
                                            Object... args)
   {
      ServerImpl server = null;
      if (h != null && h._isValid())
      {
         WrappedResource res = h.getResource();
         if (res instanceof PersistentProcedure)
         {
            handle hserver = ((PersistentProcedure) res).getServerHandle();
            
            if (!hserver.isUnknown())
            {
               server = (ServerImpl) hserver.getResource();
            }
         }
      }

      // if no server, default to SESSION
      handle hServer = (server != null ? new handle(server) : SessionUtils.asHandle());
      
      AsyncCall remoteRequest = new AsyncCall(hServer, null, name, h,
                                              asyncHandle, eventProcName, inEventprocHandle,
                                              false,
                                              modes, args)
      {
         @Override
         public void run()
         {
            // use a copy here, to not allow the user to change the reference
            invokeImpl(asyncReq, (handle) null, pHandle, pName, 
                       false, false, false, false,
                       asyncReq.getModes(), asyncReq.getArgs());
         }
      };

      invokeAsyncImpl(remoteRequest, hServer, asyncHandle, h);
   }
   
   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsync(String    name,
                                    handle    h,
                                    handle    asyncHandle, 
                                    String    eventProcName,
                                    handle    inEventprocHandle,
                                    Object... args)
   {
      invokeInAsync(new character(name), h,
                    asyncHandle, new  character(eventProcName), inEventprocHandle, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle ASYNC..  statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsync(character name,
                                    handle    h,
                                    handle    asyncHandle, 
                                    String    eventProcName,
                                    handle    inEventprocHandle,
                                    Object... args)
   {
      invokeInAsync(name, h, asyncHandle, new character(eventProcName), inEventprocHandle, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle ASYNC.. statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsync(String    name,
                                    handle    h,
                                    handle    asyncHandle, 
                                    character eventProcName,
                                    handle    inEventprocHandle,
                                    Object... args)
   {
      invokeInAsync(new character(name), h, asyncHandle, eventProcName, inEventprocHandle, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... handle IN statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsync(character name,
                                    handle    h,
                                    handle    asyncHandle, 
                                    character eventProcName,
                                    handle    inEventprocHandle,
                                    Object... args)
   {
      invokeInAsyncWithMode(name, h, 
                            asyncHandle, eventProcName, inEventprocHandle,
                            null, args);
   }
   
   /**
    * Invoke the procedure with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentWithMode(String    name,
                                               String    modes,
                                               Object... args)
   {
      invokePersistentWithMode(new character(name), modes, args);
   }

   /**
    * Invoke the procedure with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentWithMode(character name,
                                               String    modes,
                                               Object... args)
   {
      invokePersistentImpl(null, name, null, null, false, modes, args);
   }

   /**
    * Invoke the procedure with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistent(String name, Object... args)
   {
      invokePersistent(new character(name), args);
   }
   
   /**
    * Invoke the procedure with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistent(character name, Object... args)
   {
      invokePersistentImpl(null, name, null, null, false, null, args);
   }

   /**
    * Invoke the procedure with the given name, and save its procedure handle
    * in the given handle (emitted for a RUN ... PERSISTENT set handle 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code>, just adds the procedure to the
    *           persistent procedure list.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentSetWithMode(String    name,
                                                  handle    h,
                                                  String    modes,
                                                  Object... args) 
   {
      invokePersistentSetWithMode(new character(name), h, modes, args);
   }

   /**
    * Invoke the procedure with the given name, and save its procedure handle
    * in the given handle (emitted for a RUN ... PERSISTENT set handle 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code>, just adds the procedure to the
    *           persistent procedure list.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentSetWithMode(character name,
                                                  handle    h,
                                                  String    modes,
                                                  Object... args)
   {
      invokePersistentImpl(null, name, h, null, false, modes, args);
   }

   /**
    * Invoke the procedure with the given name, and save its procedure handle
    * in the given handle (emitted for a RUN ... PERSISTENT set handle 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code>, just adds the procedure to the
    *           persistent procedure list.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentSet(String    name,
                                          handle    h,
                                          Object... args) 
   {
      invokePersistentSet(new character(name), h, args);
   }

   /**
    * Invoke the procedure with the given name, and save its procedure handle
    * in the given handle (emitted for a RUN ... PERSISTENT set handle 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code>, just adds the procedure to the
    *           persistent procedure list.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentSet(character name,
                                          handle    h,
                                          Object... args)
   {
      invokePersistentImpl(null, name, h, null, false, null, args);
   }
   
   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetWithMode(String    name, 
                                                        handle    persistHandle,
                                                        handle    portHandle,
                                                        handle    remoteHandle,
                                                        boolean   transactionDistinct,
                                                        String    modes, 
                                                        Object... args) 
   {
      invokeRemotePersistentSetWithMode(new character(name), persistHandle, portHandle,
                                        remoteHandle, transactionDistinct,
                                        modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetWithMode(character name, 
                                                        handle    persistHandle,
                                                        handle    portHandle,
                                                        handle    remoteHandle,
                                                        boolean   transactionDistinct,
                                                        String    modes, 
                                                        Object... args) 
   {
      invokePersistentImpl(remoteHandle, name,
                           persistHandle, portHandle,
                           transactionDistinct, modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT SET ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSet(String    name,
                                                handle    persistHandle,
                                                handle    portHandle,
                                                handle    remoteHandle,
                                                boolean   transactionDistinct,
                                                Object... args)
   {
      invokeRemotePersistentSet(new character(name), persistHandle, portHandle,
                                remoteHandle, transactionDistinct,
                                args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT SET ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSet(character name,
                                                handle    persistHandle,
                                                handle    portHandle,
                                                handle    remoteHandle,
                                                boolean   transactionDistinct,
                                                Object... args)
   {
      invokeRemotePersistentSetWithMode(name, persistHandle, portHandle,
                                        remoteHandle, transactionDistinct,
                                        null, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT  ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentWithMode(String    name,
                                                     handle    portHandle,
                                                     handle    remoteHandle,
                                                     boolean   transactionDistinct,
                                                     String    modes, 
                                                     Object... args)
   {
      invokeRemotePersistentWithMode(new character(name), portHandle, 
                                     remoteHandle, transactionDistinct,
                                     modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT  ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentWithMode(character name,
                                                     handle    portHandle,
                                                     handle    remoteHandle,
                                                     boolean   transactionDistinct,
                                                     String    modes, 
                                                     Object... args)
   {
      invokeRemotePersistentSetWithMode(name, null, portHandle,
                                        remoteHandle, transactionDistinct,
                                        modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistent(String    name,
                                             handle    portHandle,
                                             handle    remoteHandle,
                                             boolean   transactionDistinct,
                                             Object... args)
   {
      invokeRemotePersistent(new character(name), portHandle, 
                             remoteHandle, transactionDistinct, 
                             args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistent(character name,
                                             handle    portHandle,
                                             handle    remoteHandle,
                                             boolean   transactionDistinct,
                                             Object... args)
   {
      invokeRemotePersistentWithMode(name, portHandle, 
                                     remoteHandle, transactionDistinct, 
                                     null, args);
   }
   
   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsyncWithMode(String    name, 
                                                          handle    portHandle,
                                                          handle    remoteHandle, 
                                                          boolean   transactionDistinct,
                                                          handle    asyncHandle,
                                                          String    eventProcName,
                                                          handle    inEventprocHandle,
                                                          String    modes, 
                                                          Object... args)
   {
      invokeRemotePersistentAsyncWithMode(new character(name), portHandle,
                                          remoteHandle, transactionDistinct, asyncHandle, 
                                          new character(eventProcName), inEventprocHandle,
                                          modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsyncWithMode(character name, 
                                                          handle    portHandle,
                                                          handle    remoteHandle, 
                                                          boolean   transactionDistinct,
                                                          handle    asyncHandle,
                                                          String    eventProcName,
                                                          handle    inEventprocHandle,
                                                          String    modes, 
                                                          Object... args)
   {
      invokeRemotePersistentAsyncWithMode(name, portHandle, 
                                          remoteHandle, transactionDistinct,
                                          asyncHandle, new character(eventProcName),
                                          inEventprocHandle,
                                          modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsyncWithMode(String    name, 
                                                          handle    portHandle,
                                                          handle    remoteHandle, 
                                                          boolean   transactionDistinct,
                                                          handle    asyncHandle,
                                                          character eventProcName,
                                                          handle    inEventprocHandle,
                                                          String    modes, 
                                                          Object... args)
   { 
      invokeRemotePersistentAsyncWithMode(new character(name), portHandle,
                                          remoteHandle, transactionDistinct,
                                          asyncHandle, eventProcName, inEventprocHandle,
                                          modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsyncWithMode(character name, 
                                                          handle    portHandle,
                                                          handle    remoteHandle, 
                                                          boolean   transactionDistinct,
                                                          handle    asyncHandle,
                                                          character eventProcName,
                                                          handle    inEventprocHandle,
                                                          String    modes, 
                                                          Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(name, null, portHandle,
                                             remoteHandle, transactionDistinct, 
                                             asyncHandle, eventProcName, inEventprocHandle,
                                             modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsync(String    name, 
                                                  handle    portHandle,
                                                  handle    remoteHandle, 
                                                  boolean   transactionDistinct,
                                                  handle    asyncHandle,
                                                  String    eventProcName,
                                                  handle    inEventprocHandle,
                                                  Object... args)
   {
      invokeRemotePersistentAsync(new character(name), portHandle,
                                  remoteHandle, transactionDistinct,
                                  asyncHandle, new character(eventProcName), inEventprocHandle,
                                  args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsync(character name, 
                                                  handle    portHandle,
                                                  handle    remoteHandle, 
                                                  boolean   transactionDistinct,
                                                  handle    asyncHandle,
                                                  String    eventProcName,
                                                  handle    inEventprocHandle,
                                                  Object... args)
   {
      invokeRemotePersistentAsync(name, portHandle,
                                  remoteHandle, transactionDistinct,
                                  asyncHandle, new character(eventProcName), inEventprocHandle,
                                  args);
   }
   
   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsync(String    name, 
                                                  handle    portHandle,
                                                  handle    remoteHandle, 
                                                  boolean   transactionDistinct,
                                                  handle    asyncHandle,
                                                  character eventProcName,
                                                  handle    inEventprocHandle,
                                                  Object... args)
   {
      invokeRemotePersistentAsync(new character(name), portHandle,
                                  remoteHandle, transactionDistinct,
                                  asyncHandle, eventProcName, inEventprocHandle, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsync(character name, 
                                                  handle    portHandle,
                                                  handle    remoteHandle, 
                                                  boolean   transactionDistinct,
                                                  handle    asyncHandle,
                                                  character eventProcName,
                                                  handle    inEventprocHandle,
                                                  Object... args)
   {
      invokeRemotePersistentSetAsync(name, null, portHandle,
                                     remoteHandle, transactionDistinct,
                                     asyncHandle, eventProcName, inEventprocHandle,
                                     args);
   }
   
   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET.. ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsyncWithMode(String    name,
                                                             handle    persistHandle,
                                                             handle    portHandle,
                                                             handle    remoteHandle, 
                                                             boolean   transactionDistinct,
                                                             handle    asyncHandle,
                                                             String    eventProcName,
                                                             handle    inEventprocHandle,
                                                             String    modes, 
                                                             Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(new character(name), persistHandle, portHandle,
                                             remoteHandle, transactionDistinct,
                                             asyncHandle, new character(eventProcName), 
                                             inEventprocHandle,
                                             modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET.. ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsyncWithMode(character name,
                                                             handle    persistHandle,
                                                             handle    portHandle,
                                                             handle    remoteHandle, 
                                                             boolean   transactionDistinct,
                                                             handle    asyncHandle,
                                                             String    eventProcName,
                                                             handle    inEventprocHandle,
                                                             String    modes, 
                                                             Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(name, persistHandle, portHandle, 
                                             remoteHandle, transactionDistinct,
                                             asyncHandle, new character(eventProcName), 
                                             inEventprocHandle,
                                             modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET.. ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsyncWithMode(String    name,
                                                             handle    persistHandle,
                                                             handle    portHandle,
                                                             handle    remoteHandle, 
                                                             boolean   transactionDistinct,
                                                             handle    asyncHandle,
                                                             character eventProcName,
                                                             handle    inEventprocHandle,
                                                             String    modes, 
                                                             Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(new character(name), persistHandle, portHandle,
                                             remoteHandle, transactionDistinct,
                                             asyncHandle, eventProcName, inEventprocHandle,
                                             modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET.. ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsyncWithMode(character name, 
                                                             handle    persistHandle,
                                                             handle    portHandle,
                                                             handle    remoteHandle, 
                                                             boolean   transactionDistinct,
                                                             handle    asyncHandle,
                                                             character eventProcName,
                                                             handle    inEventprocHandle,
                                                             String    modes, 
                                                             Object... args)
   {
      boolean isServer = (remoteHandle.getResource() instanceof ServerImpl);
      ServerImpl server = isServer ? (ServerImpl) remoteHandle.getResource() : null;

      if (server.isWebService())
      {
         final String err =
            "Unexpected Web service client error. Copy the following message and contact " +
            "technical support. Web service procedure not initialized in cwsSend (11461)";
         ErrorManager.recordOrThrowError(11461, err, false, true);
         return;
      }

      // do this only if the server is still connected; if not, it will fallback to non-async 
      // processing to produce the appropriate error
      final boolean obtainProxy = isServer && 
                                  server._connected() && 
                                  server.getServerHelper().isSessionFree();

      AsyncCall remoteRequest = new AsyncCall(remoteHandle, portHandle, name, new handle(),
                                              asyncHandle, eventProcName, inEventprocHandle,
                                              transactionDistinct,
                                              modes, args)
      {
         @Override
         public void run()
         {
            if (obtainProxy)
            {
               // if truly-async request, must run the "execute" method on the resolved proxy
               // if the proxy is invalid or can't map attributes or other error, then error 
               // messages are generated now
               initializeProxy(this);
            }
            else
            {
               // if not a server, go the normal way
               invokePersistentImpl(asyncReq, hServer, pName, pHandle, portHandle,
                                    transactionDistinct,
                                    asyncReq.getModes(), asyncReq.getArgs());
            }
         }
      };

      if (obtainProxy)
      {
         AppServerHelper helper = (AppServerHelper) server.getServerHelper();
         ProxyProcedureWrapper proxy = helper.obtainProxy(name);
         persistHandle.assign(proxy);
         remoteRequest.pHandle.assign(proxy);
      }
      
      invokeAsyncImpl(remoteRequest, remoteHandle, asyncHandle, new handle());
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsync(String    name,
                                                     handle    persistHandle,
                                                     handle    portHandle,
                                                     handle    remoteHandle, 
                                                     boolean   transactionDistinct,
                                                     handle    asyncHandle,
                                                     String    eventProcName,
                                                     handle    inEventprocHandle,
                                                     Object... args)
   {
      invokeRemotePersistentSetAsync(new character(name), persistHandle, portHandle, 
                                     remoteHandle, transactionDistinct,
                                     asyncHandle, new character(eventProcName), inEventprocHandle,
                                     args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsync(character name,
                                                     handle    persistHandle,
                                                     handle    portHandle,
                                                     handle    remoteHandle, 
                                                     boolean   transactionDistinct,
                                                     handle    asyncHandle,
                                                     String    eventProcName,
                                                     handle    inEventprocHandle,
                                                     Object... args)
   {
      invokeRemotePersistentSetAsync(name, persistHandle, portHandle,
                                     remoteHandle, transactionDistinct, asyncHandle, 
                                     new character(eventProcName), inEventprocHandle,
                                     args);
   }
   
   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsync(String    name, 
                                                     handle    persistHandle,
                                                     handle    portHandle,
                                                     handle    remoteHandle, 
                                                     boolean   transactionDistinct,
                                                     handle    asyncHandle,
                                                     character eventProcName,
                                                     handle    inEventprocHandle,
                                                     Object... args)
   {
      invokeRemotePersistentSetAsync(new character(name), persistHandle, portHandle,
                                     remoteHandle, transactionDistinct,
                                     asyncHandle, eventProcName, inEventprocHandle,
                                     args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsync(character name, 
                                                     handle    persistHandle,
                                                     handle    portHandle,
                                                     handle    remoteHandle, 
                                                     boolean   transactionDistinct,
                                                     handle    asyncHandle,
                                                     character eventProcName,
                                                     handle    inEventprocHandle,
                                                     Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(name, persistHandle, portHandle,
                                             remoteHandle, transactionDistinct,
                                             asyncHandle, eventProcName, inEventprocHandle,
                                             null, args);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionWithMode(Class<T>  cls,
                                 String    name,
                                 String    modes,
                                 Object... args)
   {
      BaseDataType res = invokeDynamicFunctionWithMode(new character(name),
                                                       modes,
                                                       args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionWithMode(Class<T>  cls,
                                 character name,
                                 String    modes,
                                 Object... args)
   {
      BaseDataType res = invokeDynamicFunctionInWithMode(name,
                                                         null,
                                                         modes,
                                                         args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionInWithMode(Class<T>  cls,
                                   String    name,
                                   handle    h,
                                   String    modes,
                                   Object... args)
   {
      BaseDataType res = invokeDynamicFunctionInWithMode(new character(name),
                                                         h,
                                                         modes,
                                                         args);
      return convertValue(cls, res);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionInWithMode(Class<T>  cls,
                                   character name,
                                   handle    h,
                                   String    modes,
                                   Object... args)
   {
      BaseDataType res = invokeDynamicFunctionImpl(name, h, modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunction(Class<T>  cls,
                         String    name,
                         Object... args)
   {
      BaseDataType res = invokeDynamicFunction(new character(name), args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunction(Class<T>  cls,
                         character name,
                         Object... args)
   {
      BaseDataType res = invokeDynamicFunctionIn(name, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionIn(Class<T>  cls,
                           String    name,
                           handle    h,
                           Object... args)
   {
      BaseDataType res = invokeDynamicFunctionIn(new character(name), h, args);
      return convertValue(cls, res);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionIn(Class<T>  cls,
                           character name,
                           handle    h,
                           Object... args)
   {
      BaseDataType res = invokeDynamicFunctionImpl(name, h, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionWithMode(String    name,
                                                            String    modes,
                                                            Object... args)
   {
      return invokeDynamicFunctionWithMode(new character(name), modes, args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionWithMode(character name,
                                                            String    modes,
                                                            Object... args)
   {
      return invokeDynamicFunctionInWithMode(name, 
                                             null,
                                             modes, 
                                             args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionInWithMode(String    name,
                                                              handle    h,
                                                              String    modes,
                                                              Object... args)
   {
      return invokeDynamicFunctionInWithMode(new character(name),
                                             h,
                                             modes,
                                             args);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionInWithMode(character name,
                                                              handle    h,
                                                              String    modes,
                                                              Object... args)
   {
      return invokeDynamicFunctionImpl(name, h, modes, args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunction(String    name,
                                                    Object... args)
   {
      return invokeDynamicFunction(new character(name), args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunction(character name,
                                                    Object... args)
   {
      return invokeDynamicFunctionIn(name, null, args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionIn(String    name,
                                                      handle    h,
                                                      Object... args)
   {
      return invokeDynamicFunctionIn(new character(name), h, args);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionIn(character name,
                                                      handle    h,
                                                      Object... args)
   {
      return invokeDynamicFunctionImpl(name, h, null, args);
   }

   /**
    * Emitted for plain function calls.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeFunctionWithMode(Class<T>  cls,
                          String    name,
                          String    modes,
                          Object... args)
   {
      BaseDataType res = invokeFunctionWithMode(new character(name), modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for plain function calls.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeFunctionWithMode(Class<T>  cls,
                          character name,
                          String    modes,
                          Object... args) 
   {
      BaseDataType res = invokeFunctionInWithMode(name, null, modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeFunctionInWithMode(Class<T>  cls,
                            String    name,
                            handle    h,
                            String    modes,
                            Object... args)
   {
      BaseDataType res = invokeFunctionInWithMode(new character(name),
                                                  h,
                                                  modes,
                                                  args);
      return convertValue(cls, res);
   }
   
   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeFunctionInWithMode(Class<T>  cls,
                            character name,
                            handle    h,
                            String    modes,
                            Object... args) 
   {
      BaseDataType res = invokeSimpleFunctionImpl(name, h, modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for plain function calls.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T invokeFunction(Class<T>  cls,
                                                           String    name,
                                                           Object... args)
   {
      BaseDataType res = invokeFunction(new character(name), args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for plain function calls.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T invokeFunction(Class<T>  cls,
                                                           character name,
                                                           Object... args) 
   {
      BaseDataType res = invokeFunctionIn(name, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T invokeFunctionIn(Class<T>  cls,
                                                             String    name,
                                                             handle    h, 
                                                             Object... args)
   {
      BaseDataType res = invokeFunctionIn(new character(name), h, args);
      return convertValue(cls, res);
   }
   
   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T invokeFunctionIn(Class<T>  cls,
                                                             character name,
                                                             handle    h,
                                                             Object... args) 
   {
      BaseDataType res = invokeFunctionInWithMode(name, h, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for plain function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionWithMode(String    name,
                                                     String    modes,
                                                     Object... args)
   {
      return invokeFunctionWithMode(new character(name), modes, args);
   }

   /**
    * Emitted for plain function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionWithMode(character name,
                                                     String    modes,
                                                     Object... args) 
   {
      return invokeFunctionInWithMode(name, null, modes, args);
   }

   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionInWithMode(String    name,
                                                       handle    h,
                                                       String    modes,
                                                       Object... args)
   {
      return invokeFunctionInWithMode(new character(name), h, modes, args);
   }
   
   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionInWithMode(character name,
                                                       handle    h,
                                                       String    modes,
                                                       Object... args) 
   {
      return invokeSimpleFunctionImpl(name, h, modes, args);
   }

   /**
    * Emitted for plain function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunction(String name, Object... args)
   {
      return invokeFunction(new character(name), args);
   }

   /**
    * Emitted for plain function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunction(character name, Object... args) 
   {
      return invokeFunctionIn(name, null, args);
   }

   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionIn(String    name,
                                               handle    h, 
                                               Object... args)
   {
      return invokeFunctionIn(new character(name), h, args);
   }
   
   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionIn(character name,
                                               handle    h,
                                               Object... args) 
   {
      return invokeFunctionInWithMode(name, h, null, args);
   }

   /**
    * Check if an internal procedure or function is reachable and valid. This searches the
    * specified handle and all the super-procedures.
    *
    * @param    h
    *           A valid procedure handle.
    * @param    name
    *           The procedure or function name.
    * @param    function
    *           Flag indicating if the search is for a function or for an internal procedure.
    * @param    modes
    *           The parameter modes.
    * @param    validateModes
    *           Flag indicating if the parameter modes should be validated or not.
    * @param    args
    *           The passed arguments.
    *
    * @return   null if the internal entry can't be found; otherwise ArgValidationErrors instance
    *           with error status or {@link ArgValidationErrors#NO_ERROR} if error is not exists.
    *
    */
   public static ArgValidationErrors reachableInternalEntry(handle    h,
                                                            String    name,
                                                            boolean   function,
                                                            String    modes,
                                                            boolean   validateModes,
                                                            Object... args)
   {
      WorkArea wa = work.get();
      InternalEntryCaller caller = null;
      for (Resolver r : wa.inHandleResolvers)
      {
         caller = r.resolve(h, name, function, false, null);

         if (caller != null)
         {
            break;
         }
      }

      if (caller == null)
      {
         // was not found
         return null;
      }

      ArgValidationErrors valid = caller.valid(function, args);
      if (valid != ArgValidationErrors.NO_ERROR)
      {
         return valid;
      }

      // validate the parameter modes
      String definedModes = caller.getParameterModes(function);
      if (validateModes && 
          (modes != null && !modes.equals(definedModes) ||
           (modes == null && definedModes != null)))
      {
         return ArgValidationErrors.INCORRECT_TYPES;
      }

      return valid;
   }

   /**
    * The implementation of all function and procedure calls. This disambiguates and implements
    * the following calls:
    * <ul>
    *    <li>if <code>h</code> is not null, then this is a RUN ... IN handle call.</li>
    *    <li>if the <code>exports</code> is not null, then this is a RUN ... ON SERVER call, so
    *        only external programs will be searched.  Note that this parameter will be not-null
    *        only when this call is executed on the remote, appserver, side.</li>
    *    <li>else, this is a RUN statement call with no special clauses.</li>
    * </ul>
    * 
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code>, defaults to
    *           THIS-PROCEDURE.
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which is the <code>unknown
    *           value</code> by default, on error or for procedures.
    */
   static BaseDataType invoke(handle    h,
                              character name,
                              boolean   function,
                              boolean   dynamicFunction,
                              boolean   superCall,
                              boolean   transactionDistinct,
                              String    modes,
                              Object... args)
   {
      // h will never be a proxy procedure handle at the time of this call. if it ends up a
      // remote procedure handle here, we have a problem in the runtime implementation
      if (ProcedureManager.isProxy(h))
      {
         throw new IllegalArgumentException("The supplied procedure handle is a proxy handle!");
      }
      
      WorkArea wa = work.obtain();
      List<Resolver> resolvers = null;
      if (h != null)
      {
         // The target of a "RUN proc" stmt or DYNAMIC-FUNCTION call ends on 
         // first find in: 
         // - THIS-PROCEDURE 
         // - THIS-PROCEDURE:super-procedures 
         // - SESSION-super-procedures
         resolvers = wa.inHandleResolvers;
      }
      else
      {
         // The target of a "RUN proc" stmt or DYNAMIC-FUNCTION call ends on 
         // first find in: 
         // - THIS-PROCEDURE 
         // - THIS-PROCEDURE:super-procedures 
         // - SESSION-super-procedures
         // - if an external program with the given name is found (for the 
         // "RUN proc" statement without "IN handle" clause)
         // - search distinguishes between procedures and functions with the 
         // same name. On first name match, will attempt to invoke it. 
         // - is not necessarily for the target procedure to have its header 
         // defined in the current procedure. Applies only to "RUN proc" 
         // statement. 
         // - "RUN proc0." and "proc0()" will end up calling different code 
         // (one is a super-proc, the other is a super-function). 
         // The body for these two may be in different super-procedures.
         resolvers = wa.resolvers;
      }

      return invokeImpl(resolvers,
                        h, 
                        name, 
                        function, 
                        dynamicFunction, 
                        superCall,
                        transactionDistinct,
                        modes,
                        args);
   }

   /**
    * Resolve the external program with the specified name.
    * 
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    exports
    *           The list of allowed external procedures to be invoked with a RUN ... ON SERVER
    *           statement. 
    * 
    * @return  The instantiated (not yet initialized) {@link ExternalProgramWrapper}.  
    */
   static ExternalProgramWrapper resolveExternalProcedure(String name, character exports)
   {
      InternalEntryCaller caller = work.obtain().externalResolver
                                                .resolve(null, name, false, false, exports);
      
      ExternalProgramWrapper extProg = null;
      
      if (caller != null)
      {
         extProg = new ExternalProgramWrapper(caller.getCallerInstance());
         
         ProcedureManager.addProcedure(new handle(extProg), name, true);
      }
      
      return extProg;
   }

   /**
    * Initialize the specified external procedure. This is done by calling the associated
    * <code>execute</code> method.
    * 
    * @param    extProg
    *           The associated external program.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    */
   static void initializeExternalProcedure(ExternalProgramWrapper extProg,
                                           boolean                transactionDistinct,
                                           String                 modes,
                                           Object...              args)
   {
      InternalEntryCaller caller = new ExternalProcedureCaller(extProg.get());

      // resolve the name early, as at the time the invokeError is called the proc handle is gone
      String relName = ProcedureManager.getRelativeName(extProg.referent);

      Exception deferred = null;
      handle hProc = new handle(extProg);
      try
      {
         if (!validArguments(relName, caller, false, modes, args))
         {
            return;
         }
   
         boolean error = false;
         try
         {
            Object referent = extProg.get();
            
            ProcedureManager.pushCalleeInfo(false, referent, referent, 
                                            ProcedureManager.EXTERNAL_PROGRAM, relName,
                                            false, true);
            
            caller.invoke(modes, args);
         }
         catch (Exception e)
         {
            // any exceptions here cause the persistent procedure to no longer be available
            error = true;
   
            throw e;
         }
         finally
         {
            ProcedureManager.popCalleeInfo();
            
            if (error || ErrorManager.isPending())
            {
               ProcedureManager.removePersistentProcedure(hProc);
            }
         }
      }
   
      // allow Progress compatible exceptions to flow through
      catch (ConditionException ce)
      {
         throw ce;
      }
      
      catch (RetryUnwindException rue)
      {
         throw rue;
      }
      
      // handle reflection exceptions
      catch (NoSuchMethodException nsm)
      {
         invokeFailure(caller.getMethodName());
         return;
      }
      
      catch (InvocationTargetException exc)
      {
         // if the exception thrown was a P2J runtime type, we
         // must handle it normally, so extract the root cause
         // exception and rethrow it
         Throwable cause = exc.getCause();
         if (cause != null)
         {
            if (cause instanceof ConditionException)
            {
               throw (ConditionException) cause;
            }
            if (cause instanceof RetryUnwindException)
            {
               throw (RetryUnwindException) cause;
            }
            
            // exception is a runtime problem that we don't
            // handle here
            deferred = exc;
         }
      }
      
      // all other cases
      catch (Exception exp)
      {
         deferred = exp;
      }
      
      if (deferred != null)
      {
         invokeError(caller, relName, deferred);
      }
   }
   
   /**
    * The implementation of all RUN ... PERSISTENT calls or appserver calls for an external
    * procedure. If the <code>exports</code> parameter is not null, then the ON SERVER clause is
    * in effect; this parameter can be not-null only on the remote, appserver, side.
    * 
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    persistent
    *           Flag indicating that this call needs to be persistent.
    * @param    h
    *           The handle where to save the persistent procedure.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    exports
    *           The list of allowed external procedures to be invoked with a RUN ... ON SERVER
    *           statement. 
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    */
   static void invokeExternalProcedure(character name,
                                       boolean   persistent,
                                       handle    h,
                                       boolean   transactionDistinct,
                                       character exports,
                                       String    modes,
                                       Object... args)
   {
      String pname = name.toStringMessage();
      
      InternalEntryCaller caller = null;
      Exception deferred = null;
      try
      {
         caller = work.obtain().externalResolver.resolve(null, pname, false, false, exports);
         
         if (caller == null)
         {
            invokeFailure(pname);
            return;
         }
   
         if (!validArguments(pname, caller, false, modes, args))
         {
            return;
         }
   
         Object instance = caller.getCallerInstance();
         if (h == null)
         {
            h = new handle();
         }
         h.set(new ExternalProgramWrapper(instance));
   
         boolean error = false;
         try
         {
            ProcedureManager.pushCalleeInfo(false, h.get(), h.get(),
                                            ProcedureManager.EXTERNAL_PROGRAM, pname, 
                                            false, persistent);
            BlockManager.markPendingDynCall();
            
            caller.invoke(modes, args);
         }
         catch (Exception e)
         {
            error = true;
   
            throw e;
         }
         finally
         {
            ProcedureManager.popCalleeInfo();
            
            if (persistent && (error || ErrorManager.isPending()))
            {
               // any exceptions here cause the persistent procedure to no longer
               // be available
               ProcedureManager.removePersistentProcedure(h);
            }
         }
      }
   
      // allow Progress compatible exceptions to flow through
      catch (ConditionException ce)
      {
         throw ce;
      }
      
      catch (RetryUnwindException rue)
      {
         throw rue;
      }
      
      // handle reflection exceptions
      catch (NoSuchMethodException nsm)
      {
         invokeFailure(caller.getMethodName());
         return;
      }
      
      catch (InvocationTargetException exc)
      {
         // if the exception thrown was a P2J runtime type, we
         // must handle it normally, so extract the root cause
         // exception and rethrow it
         Throwable cause = exc.getCause();
         if (cause != null)
         {
            if (cause instanceof ConditionException)
            {
               throw (ConditionException) cause;
            }
            if (cause instanceof RetryUnwindException)
            {
               throw (RetryUnwindException) cause;
            }
            
            // exception is a runtime problem that we don't
            // handle here
            deferred = exc;
         }
      }
      
      // all other cases
      catch (Exception exp)
      {
         deferred = exp;
      }
      
      if (deferred != null)
      {
         invokeError(caller, pname, deferred);
      }
   }

   /**
    * Convert BaseDataType value to instance of given class. If value is <code>null</code>,
    * returns <code>null</code>. If value is <code>unknown value</code>, creates a new instance
    * of the given type that represents the <code>unknown value</code>, otherwise returns value
    * casted to given type.
    *
    * @param    cls
    *           Class.
    * @param    value
    *           Value.
    *
    * @return   If value is <code>null</code>, returns <code>null</code>. If value is
    *           <code>unknown value</code>, creates a new instance of the given type that
    *           represents the <code>unknown value</code>, otherwise returns value casted to
    *           given type.
    *
    * @throws   IllegalStateException
    *           If value is not compatible with given class.
    */
   private static <T extends BaseDataType> T convertValue(Class<T> cls, BaseDataType value)
   {
      if (value == null)
      {
         return null;
      }

      if (value instanceof unknown)
      {
         value = BaseDataType.generateUnknown(cls);
      }

      if (!value.getClass().isAssignableFrom(cls))
      {
         throw new IllegalStateException("Value of type " + value.getClass().getName() +
               " is not compatible with expected return type " + cls.getName());
      }

      return (T) value;
   }

   /**
    * Execute an async procedure call.
    * 
    * @param    request
    *           The request which will execute the procedure call.
    * @param    hServer
    *           The server used to execute this async call. 
    * @param    asyncHandle
    *           The handle to the asynchronous request.     * @param    pHandle
    *           The handle to which the procedure belongs, if it is RUN ... IN handle statement.
    */
   private static void invokeAsyncImpl(AsyncCall request,
                                       handle    hServer,
                                       handle    asyncHandle,
                                       handle    pHandle)
   {
      // expose the async resource to the caller
      asyncHandle.assign(request.asyncReq);

      boolean isServer = (hServer.getResource() instanceof ServerImpl);
      ServerImpl server = isServer ? (ServerImpl) hServer.getResource() : null;

      // register for notifications the server and the procedure handle
      if (isServer)
      {
         request.asyncReq.registerListener(server);
      }
      // check the procedure handle
      if (pHandle._isValid() && pHandle.getResource() instanceof AsyncRequestListener)
      {
         request.asyncReq.registerListener((AsyncRequestListener) pHandle.getResource());
      }
      
      // notify all listeners an async call is in effect
      request.asyncReq.notifyStart();
      
      // perform invocation
      if (isServer && server._connected())
      {
         // let the server run this async request
         server.invokeAsync(request.asyncReq);
      }
      else
      {
         // invoke immediately
         request.asyncReq.execute(false);
      }
   }

   /**
    * The implementation for all dynamic function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   private static BaseDataType invokeDynamicFunctionImpl(character name,
                                                         handle    h,
                                                         String    modes,
                                                         Object... args) 
   {
      return invokeFunctionImpl(name, h, true, false, modes, args);
   }

   /**
    * The implementation for all "simple" function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   private static BaseDataType invokeSimpleFunctionImpl(character name,
                                                        handle    h,
                                                        String    modes,
                                                        Object... args) 
   {
      return invokeFunctionImpl(name, h, false, false, modes, args);
   }

   /**
    * The implementation for all function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a SUPER() call.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   private static BaseDataType invokeFunctionImpl(character name,
                                                  handle    h,
                                                  boolean   dynamicFunction,
                                                  boolean   superCall,
                                                  String    modes,
                                                  Object... args) 
   {
      // TODO: add support for functions returning an extent
      return invokeImpl((handle) null, h, name, 
                        true, dynamicFunction, superCall,
                        false, modes, args);
   }

   /**
    * The implementation for RUN SUPER and RUN ... IN handle calls.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER call.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    */
   private static void invokeInImpl(character name, 
                                    handle    h,
                                    boolean   superCall,
                                    String    modes,
                                    Object... args)
   {
      invokeImpl((handle) null, h, name, false, false, superCall, false, modes, args);
   }
   
   
   /**
    * The implementation of function or procedure calls which do not use the
    * IN handle clause.
    * 
    * @param    server
    *           The target appserver, to which this call will be dispatched.
    *           Is <code>null</code> if the RUN statement doesn't have the ON SERVER clause.
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code> defaults to 
    *           THIS-PROCEDURE. If this is a proxy procedure, the call is dispatched to the
    *           procedure's server.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default, on error or for
    *           procedures.
    */
   private static BaseDataType invokeImpl(handle    server,
                                          handle    h,
                                          character name,
                                          boolean   function,
                                          boolean   dynamicFunction,
                                          boolean   superCall,
                                          boolean   transactionDistinct,
                                          String    modes,
                                          Object... args)
   {
      return invokeImpl(null, server, h, name,
                        function, dynamicFunction, superCall, transactionDistinct,
                        modes, args);
   }

   /**
    * The implementation of function or procedure calls which do not use the
    * IN handle clause.
    * 
    * @param    asyncReq
    *           When not-null, an {@link AsyncRequestImpl} instance for the associated async
    *           request.
    * @param    server
    *           The target appserver, to which this call will be dispatched.
    *           Is <code>null</code> if the RUN statement doesn't have the ON SERVER clause.
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code> defaults to 
    *           THIS-PROCEDURE. If this is a proxy procedure, the call is dispatched to the
    *           procedure's server.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default, on error or for
    *           procedures.
    */
   private static BaseDataType invokeImpl(AsyncRequestImpl asyncReq,
                                          handle           server,
                                          handle           h,
                                          character        name,
                                          boolean          function,
                                          boolean          dynamicFunction,
                                          boolean          superCall,
                                          boolean          transactionDistinct,
                                          String           modes,
                                          Object...        args)
   {
      // save the original received server instance.
      final handle origServer = server;

      if (server == null && h != null && h._isValid())
      {
         WrappedResource res = h.getResource();
         
         // get the server from the handle
         if (ProcedureManager.isProxy(h))
         {
            server = new handle(((ProxyProcedureWrapper) res).getServer());
         }
         else if (res instanceof PortTypeWrapper)
         {
            server = new handle(((PortTypeWrapper) res).getServer());
         }
      }
      
      // if the server is the SESSION handle, then this is a local call
      if (server == null || CompareOps._isEqual(server, SessionUtils.asHandle()))
      {
         // this is a normal call, invoke immediately
         return invoke(h, name, function, dynamicFunction, superCall, false, modes, args);
      }
      else
      {
         if (!validServer(server, h, name))
         {
            return new unknown();
         }

         ServerImpl s = (ServerImpl) server.get();
         boolean isWebService = s.isWebService();
         
         if (isWebService && origServer != null && h == null)
         {
            // the target portHandle must always be specified in case of 
            // RUN port-type ON SERVER h-web-server statements.
            // if is missing, the RUN statement is a no-op
            
            return new unknown();
         }

         BaseDataType result = new unknown();
         try
         {
            // if sync mode, no async requests must be running
            if (asyncReq == null && s.hasAsyncRequests())
            {
               return new unknown();
            }

            // invoke appServer
            ServerHelper helper = s.getServerHelper();
            // if ON SERVER is used (h is null), only external procedures can be invoked.
            // this is managed by the remote side.
            result = helper.invoke(asyncReq, name, h,
                                   function, dynamicFunction, superCall,
                                   transactionDistinct, modes, args);
            
            if (!function && asyncReq == null && result != null && !isWebService)
            {
               // TODO: the return-value is set for web-service requests only if the web-service 
               // is configured and exported via Progress.

               setReturnValue((character) result);
               result = new unknown();
            }
         }
         catch (ConditionException e)
         {
            if (!isWebService && e.getCause() instanceof NumberedException)
            {
               ErrorManager.recordOrThrowError((NumberedException) e.getCause());
            }
            else
            {
               throw e;
            }
         }

         return result;
      }
   }
   
   /**
    * The implementation of all function and procedure calls.
    * 
    * @param    resolvers
    *           The {@link Resolver} instances to use to resolve the internal
    *           entry to a Java method, based on the passed parameters.
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    cname
    *           The legacy 4GL name for the function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default, on error or for
    *           procedures.
    */
   private static BaseDataType invokeImpl(List<Resolver> resolvers,
                                          handle         h,
                                          character      cname,
                                          boolean        function,
                                          boolean        dynamicFunction,
                                          boolean        superCall,
                                          boolean        transactionDistinct,
                                          String         modes,
                                          Object...      args)
   {
      BaseDataType result = new unknown();

      String name = cname.toStringMessage();

      // handle validation is done first
      boolean explicit = h != null;
      if (!explicit)
      {
         h = ProcedureManager.thisProcedure();
      }
     
      // this is a special case only for functions. if the handle has been
      // deleted and it still contains the reference to the procedure, then
      // it defaults to this-procedure
      if (function && !h._isValid() && ProcedureManager.isProcedure(h))
      {
         h = ProcedureManager.thisProcedure();
      }

      if (!validHandle(h, name, explicit, function, dynamicFunction))
      {
         return result;
      }
   
      if (cname.isUnknown() || cname.toStringMessage().length() == 0)
      {
         String err = null;
         int errid = -1;

         if (function && dynamicFunction)
         {
            err = "Could not invoke dynamic user-defined function";
            errid = 5640;
         }
         
         if (err == null)
         {
            if (cname.isUnknown())
            {
               invokeFailure(name);
               return result;
            }
            else
            {
               err = "No PROGRESS program specified for RUN or COMPILE";
               errid = 467;
            }
         }
         
         ErrorManager.recordOrThrowError(errid, err);
         return result;
      }

      Exception deferred = null;
      InternalEntryCaller caller = null;
      try
      {
         handle hTarget = new handle(h);

         for (Resolver r : resolvers)
         {
            caller = r.resolve(h, name, function, superCall, null);
      
            if (caller instanceof MapToNotFound)
            {
               // if a map-to was searched and not explicitly found, search in super-procedures
               name = caller.getInternalEntryName();
               caller = null;
            }

            if (caller != null)
            {
               break;
            }
         }
         
         if (caller == null)
         {
            String pname = ProcedureManager.getRelativeName(h.get());
            if (function)
            {
               // if the function is defined but could not be resolved, 
               // then show error
               if (dynamicFunction)
               {
                  final String msg = 
                     "User-defined function '%s' invoked  dynamically but could not be found";
                  if (SessionUtils._isRemote())
                  {
                     ErrorManager.recordOrThrowError(5639, String.format(msg, name), false);
                  }
                  else
                  {
                     ErrorManager.displayError(5639, String.format(msg, name), false);
                  }
               }
               else
               {
                  if (SessionUtils._isRemote())
                  {
                     final String msg =
                        "User-defined function '%s' invoked but could not be found";
                     ErrorManager.recordOrThrowError(6445, String.format(msg, name), false);
                  }
                  else
                  {
                     final String msg = "External user defined function '%s' is not defined in " +
                                        "the procedure context given";
                     ErrorManager.recordOrShowError(2779, String.format(msg, name),
                                                    false, false, false);
                  }
               }
               
               return result;
            }
            else if (explicit)
            {
               final String msg = "Procedure %s has no entry point for %s";
               ErrorManager.recordOrThrowError(6456, 
                                               String.format(msg, pname, name),
                                               false);
               return result;
            }

            invokeFailure(name);
            return result;
         }
         else
         {
            // caller.phandle is null in cases of external programs
            handle callerHandle = caller.getPhandle();
            
            if (callerHandle != null && !CompareOps._isEqual(callerHandle, h))
            {
               h = callerHandle;
               
               // TODO: this should recursively call invokeImpl(server, handle, ....), as the
               // new handle might be for a remote procedure
               
               // handle has switched, validate the new one. dynamic-function
               // needs to be false here.
               if (!validHandle(h, name, explicit, function, false))
               {
                  return result;
               }
            }
         }
         
         if (function)
         {
            // this recursive call is needed only for functions, as only they
            // can be set as IN handle.
            String progName = ProcedureManager.getAbsoluteName(h.get());
            if (SourceNameMapper.isInHandle(progName, name, function))
            {
               try
               {
                  return invokeImpl(resolvers,
                                    h,
                                    cname,
                                    function,
                                    dynamicFunction,
                                    superCall,
                                    transactionDistinct,
                                    modes,
                                    args);
               }
               catch (StackOverflowError t)
               {
                  final String msg =
                     "Giving up search for external user defined function " +
                     "'%s' after 64 indirections";
                  String err = String.format(msg, name);
                  int num = 2757;

                  ErrorManager.displayError(num, err, false);
                  
                  return result;
               }
            }
         }
      
         if (!validArguments(name, caller, function, modes, args))
         {
            return result;
         }

         boolean external = (caller instanceof ExternalProcedureCaller);

         try
         {
            final String iename;
            final String pname;
            if (external)
            {
               // we are invoking an external program. the target-procedure will be the 
               // to-be-invoked procedure.
               ExternalProgramWrapper extProg = new ExternalProgramWrapper(caller.callerInstance);
               hTarget = new handle(extProg);
               
               iename = ProcedureManager.EXTERNAL_PROGRAM;
               pname = name;
            }
            else
            {
               iename = name;
               pname = null;
            }

            ProcedureManager.pushCalleeInfo(superCall, hTarget.get(), caller.callerInstance, 
                                            iename, pname, function, false);
            BlockManager.markPendingDynCall();

            Object res = caller.invoke(modes, args);
            if (res == null)
            {
               if (!function)
               {
                  result = new unknown();
               }
               else
               {
                  final String msg = "Function call '%s' has returned a null value!";
                  throw new NullPointerException(String.format(msg, name));
               }
            }
            else if (res instanceof BaseDataType)
            {
               result = (BaseDataType) res;
            }
            else
            {
               if (function)
               {
                  final String spec = 
                     "Incompatible return value by function %s: [%s] of " +
                     "type [%s]";
                  String err = String.format(spec, name,
                                             result.toString(), 
                                             result.getClass().toString());
                  ErrorManager.displayError(err);

                  throw new StopConditionException(err);
               }
            }
         }
         
         finally
         {
            ProcedureManager.popCalleeInfo();
         }
      }
   
      // allow Progress compatible exceptions to flow through
      catch (ConditionException ce)
      {
         throw ce;
      }
      
      catch (RetryUnwindException rue)
      {
         throw rue;
      }
      
      // handle reflection exceptions
      catch (NoSuchMethodException nsm)
      {
         // should never happen, unless name_map.xml is corrupt
         deferred = nsm;
      }
      
      catch (InvocationTargetException exc)
      {
         // if the exception thrown was a P2J runtime type, we
         // must handle it normally, so extract the root cause
         // exception and rethrow it
         Throwable cause = exc.getCause();
         while (cause != null)
         {
            if (cause instanceof ConditionException)
            {
               throw (ConditionException) cause;
            }
            if (cause instanceof RetryUnwindException)
            {
               throw (RetryUnwindException) cause;
            }
            
            // exception is a runtime problem that we don't
            // handle here
            if (deferred == null)
            {
               deferred = exc;
            }

            cause = cause.getCause();
         }
      }
      
      // all other cases
      catch (Exception exp)
      {
         deferred = exp;
      }
      
      if (deferred != null)
      {
         invokeError(caller, name, deferred);
      }
      
      return result;
   }

   /**
    * The implementation for the RUN ... PERSISTENT [SET handle] [ON SERVER server] statement. 
    * 
    * @param    server
    *           The server instance on which to run the procedure.  This will save the procedure
    *           in the remote persistent procedure list, not on the requester side.  May be null.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code> and not ran remotely, just adds the procedure to the
    *           persistent procedure list.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   private static void invokePersistentImpl(handle    server,
                                            character name,
                                            handle    persistHandle,
                                            handle    portHandle,
                                            boolean   transactionDistinct,
                                            String    modes,
                                            Object... args) 
   {
      invokePersistentImpl(null, server, name,
                           persistHandle, portHandle,
                           transactionDistinct,
                           modes, args);
   }
   
   /**
    * Initialize the proxy procedure associated with the async request.
    * 
    * @param    call
    *           The async request which currently holds a not-initialized proxy procedure.
    */
   private static void initializeProxy(AsyncCall call)
   {
      if (!validServer(call.hServer, call.pHandle, call.pName))
      {
         return;
      }

      ServerImpl s = (ServerImpl) call.hServer.get();
      if (s.isWebService())
      {
         throw new IllegalStateException(
                                "initializeProxy can be used only with an appserver connection!");
      }
      else
      {
         try
         {
            // execute the proxy
            AppServerHelper appServer = (AppServerHelper) s.getServerHelper();
            appServer.initializeProxy(call.asyncReq,
                                      (String) call.pHandle.get(), 
                                      call.transactionDistinct,
                                      call.asyncReq.getModes(),
                                      call.asyncReq.getArgs());
         }
         catch (ConditionException e)
         {
            if (e.getCause() instanceof NumberedException)
            {
               ErrorManager.recordOrThrowError((NumberedException) e.getCause());
            }
            else
            {
               throw e;
            }
         }
         finally
         {
            call.asyncReq.complete();
         }
      }
   }
   
   /**
    * The implementation for the RUN ... PERSISTENT [SET handle] [ON SERVER server] statement. 
    * 
    * @param    asyncReq
    *           When not-null, an {@link AsyncRequestImpl} instance for the associated async
    *           request.
    * @param    server
    *           The server instance on which to run the procedure.  This will save the procedure
    *           in the remote persistent procedure list, not on the requester side.  May be null.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code> and not ran remotely, just adds the procedure to the
    *           persistent procedure list.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   private static void invokePersistentImpl(AsyncRequestImpl asyncReq,
                                            handle           server,
                                            character        name,
                                            handle           persistHandle,
                                            handle           portHandle,
                                            boolean          transactionDistinct,
                                            String           modes,
                                            Object...        args) 
   {
      // this will always be an external procedure call
      if (name.isUnknown() || name.toStringMessage().length() == 0)
      {
         String err = "No PROGRESS program specified for RUN or COMPILE";
         ErrorManager.recordOrThrowError(467, err);
         return;
      }

      // if the server is the SESSION handle, then this is a local call
      if (server == null || CompareOps._isEqual(server, SessionUtils.asHandle()))
      {
         // this is a normal call, invoke immediately
         invokeExternalProcedure(name, true, persistHandle, false, (character) null, modes, args);
      }
      else
      {
         // validate server first
         if (!validServer(server, null, name))
         {
            return;
         }

         ServerImpl s = (ServerImpl) server.get();

         if (s.isWebService())
         {
            if (portHandle == null)
            {
               final String err =
                  "The PERSISTENT keyword is invalid when used with a SERVER handle whose " +
                  "SUBTYPE is 'WEBSERVICE'";
               ErrorManager.recordOrThrowError(11481, err, false);
               return;
            }
            
            if (asyncReq != null)
            {
               final String err =
                  "Unexpected Web service client error. Copy the following message and contact " +
                  "technical support. Web service procedure not initialized in cwsSend";
               ErrorManager.recordOrThrowError(11461, err, false);
               return;
            }

            // this will establish a port connection
            WebServiceHelper helper = (WebServiceHelper) s.getServerHelper();
            PersistentProcedure portProcedure = helper.runPort(name);
            if (portProcedure != null)
            {
               // the port was ran, assign it
               portHandle.assign(portProcedure);
               
               // this is not a persistent procedure, and does not get registered with the 
               // ProcedureManager
            }
            
            return;
         }
         else
         {
            if (portHandle != null)
            {
               final String err = 
                  "'RUN ... ON SERVER server-handle' with the [SET handle] option is not valid " +
                  "without the PERSISTENT keyword unless the server is a Web service";
               ErrorManager.recordOrThrowError(14587, err, false);
               return;
            }

            // if sync mode, no async requests must be running
            if (asyncReq == null && s.hasAsyncRequests())
            {
               return;
            }

            ProxyProcedureWrapper proxyProcedure = null;
            handle h = persistHandle;

            AppServerHelper appServer = (AppServerHelper) s.getServerHelper();
            try
            {
               if (h == null)
               {
                  h = new handle();
               }
               character result = appServer.invokePersistent(asyncReq, name, h,
                                                             transactionDistinct,
                                                             modes, args);
               if (asyncReq == null)
               {
                  setReturnValue(result);
               }
               
               proxyProcedure = (ProxyProcedureWrapper) h.getResource();
               if (persistHandle != null)
               {
                  persistHandle.set(proxyProcedure);
               }
               proxyProcedure.setServer(s);
            }
            catch (ConditionException e)
            {
               if (e.getCause() instanceof NumberedException)
               {
                  ErrorManager.recordOrThrowError((NumberedException) e.getCause());
               }
               else
               {
                  throw e;
               }
            }

            // chain the procedure
            handle last = s.lastProcedure();
            if (last.isUnknown())
            {
               s.setFirstProcedure(proxyProcedure);
               s.setLastProcedure(proxyProcedure);
            }
            else
            {
               ProxyProcedureWrapper lr = (ProxyProcedureWrapper) last.getResource();
               lr.setNextSibling(proxyProcedure);
               proxyProcedure.setPrevSibling(lr);
               s.setLastProcedure(proxyProcedure);
            }
         }
      }
   }
   
   /**
    * Check if the caller can validate the arguments. If not, show an 
    * appropriate error message and throw error condition.
    * 
    * @param    caller
    *           The caller used to validate the arguments.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The argument list.
    *           
    * @return   <code>true</code> if the arguments are valid.
    */
   private static boolean validArguments(String              name,
                                         InternalEntryCaller caller,
                                         boolean             function,
                                         String              modes,
                                         Object...           args)
   {
      ArgValidationErrors argsValid = caller.valid(function, args);
      
      // deferred validation needs no further processing here
      if (argsValid == ArgValidationErrors.DEFER_VALIDATION)
      {
         return true;
      }
      
      if (caller.iename != null)
      {
         // we have an internal entry, determine the relative name used to start this referent
         name = ProcedureManager.getRelativeName(caller.callerInstance);
      }
      
      if (argsValid == ArgValidationErrors.NO_ERROR)
      {
         // validate the parameter's modes
         if (modes != null)
         {
            String definedModes = caller.getParameterModes(function);
            if (!modes.equals(definedModes))
            {
               final String errp = 
                  "Mismatched parameters types passed to procedure %s %s";
               final String errf =
                  "Mismatched parameter INPUT/OUTPUT mode passed to " +
                  "routine %s %s";
               
               String err = function ? errf : errp;
               int errno = function ? 6349 : 3230;
               String msg = String.format(err, caller.getInternalEntryName(), name);
               ErrorManager.recordOrThrowError(errno, msg, false);
               return false;
            }
         }
         
         return true;
      }

      String[] errs = null;
      int[] errIds = null;
      
      character thisName = new character();
      if (SessionUtils._isRemote())
      {
         thisName.assign("(Remote client)");
      }
      else
      {
         handle pthis = ProcedureManager.thisProcedure();
         thisName.assign(ProcedureManager.getRelativeName(pthis.get()));
      }

      switch (argsValid)
      {
         case UNEXPECTED_PARAM:
            errIds = new int[] { 1005 };
            String msg = "Procedure %s passed parameters to %s, which did not expect any";
            msg = String.format(msg,  thisName.toStringMessage(), name);
            errs = new String[] { msg };
            break;
         case INCORRECT_COUNT:
            errIds = new int[] { 3234 };
            msg = "Mismatched number of parameters passed to routine %s %s";
            msg = String.format(msg, caller.getInternalEntryName(), name);
            errs = new String[] { msg };
            break;
         case INCORRECT_TYPES:
            errIds = new int[] { 5729, 2570 };
            msg = "Routine %s sent called routine %s %s mismatched parameters";
            msg = String.format(msg, 
                                thisName.toStringMessage(),
                                caller.getInternalEntryName(),
                                name);
            
            errs = new String[]
            {
               "Incompatible datatypes found during runtime conversion",
               msg
            };
            break;
         case ARRAY_FOR_NON_ARRAY:
         case NON_ARRAY_FOR_ARRAY:
            errIds = new int[] { 11428 };
            // TODO: msg doesn't end in DOT in 4GL
            msg = "Extent parameter dimension of %d from procedure %s is mismatched with " +
                  "sub-procedure %s %s parameter dimension of %d ...(328)";
            msg = String.format(msg,
                                argsValid.paramExtent,
                                thisName.toStringMessage(),
                                caller.getInternalEntryName(),
                                name,
                                argsValid.expectedExtent);
            errs = new String[] { msg };
            break;
         case DYN_ARRAY_FOR_NON_ARRAY:
            errIds = new int[] { 12290 };
            msg = "Calling procedure %s cannot pass an indeterminate extent parameter if the " +
                  "sub-procedure %s %s parameter is a non-array parameter";
            msg = String.format(msg,
                                thisName.toStringMessage(),
                                caller.getInternalEntryName(),
                                name);
            errs = new String[] { msg };
            break;
         case NON_ARRAY_FOR_DYN_ARRAY:
            errIds = new int[] { 15472 };
            msg = 
               "Routine %s passed a non-array parameter to an array parameter of routine %s %s";
            msg = String.format(msg,
                                thisName.toStringMessage(),
                                caller.getInternalEntryName(),
                                name);
            errs = new String[] { msg };
            break;
      }
      
      ErrorManager.recordOrThrowError(errIds, errs, false);
      
      return false;
   }
   
   /**
    * Check if the handle is valid for function or procedure invocation.
    * If is invalid, show an appropriate message.
    * 
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    explicit
    *           <code>true</code> if the passed handle was explicitly set (for
    *           RUN ... IN or FUNCTION ... IN handle cases.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    *           
    * @return   <code>true</code> if the handle is valid for function or
    *           procedure invocation.
    */
   private static boolean validHandle(handle  h,
                                      String  name,
                                      boolean explicit,
                                      boolean function,
                                      boolean dynamicFunction)
   {
      if (function)
      {
         boolean isProcedure = ProcedureManager.isProcedure(h);
         if (h.isUnknown() || !isProcedure)
         {
            String err = null;
            int num = -1;

            if (!h.isUnknown() && !isProcedure)
            {
               final String msg =
                  "Could not locate external function '%s'. The handle to " +
                  "the procedure context is invalid";
               err = String.format(msg, name);
               num = 2777;
            }
            else if (explicit && dynamicFunction)
            {
               final String msg =
                  "DYNAMIC-FUNCTION function given invalid or Unknown " +
                  "context to run IN";
               err = msg;
               num = 5705;
            }
            else
            {
               final String msg =
                  "Could not evaluate the expression describing the " +
                  "context of external function '%s'";
               err = String.format(msg, name);
               num = 2767;
            }

            ErrorManager.displayError(num, err, false);
            
            return false;
         }
      }
      else
      {
         if (explicit && h.isUnknown())
         {
            String err = null;
            int num = -1;
            final String msg =
               "Could not evaluate procedure handle expression while " +
               "trying to execute %s";
             err = String.format(msg, name);
             num = 2125;
             ErrorManager.recordOrThrowError(num, err, false);
            
            return false;
         }
         if (!ProcedureManager.hasReferent(h.get()))
         {
            String err = null;
            int num = -1;
            final String msg = 
               "Invalid or inappropriate handle value given to RUN...IN " +
               "statement. Procedure '%s':%d";
            num = 2128;
            
            handle pthis = ProcedureManager.thisProcedure();
            String thisName = ProcedureManager.getRelativeName(pthis.get());
            err = String.format(msg, thisName, 0);
            ErrorManager.recordOrThrowError(num, err, false);

            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Check if the given handle is a valid, connected appserver.
    * 
    * @param    hServer
    *           The handle to validate as an appserver.
    * @param    hProc
    *           A procedure handle for IN handle cases.
    * @param    name
    *           The target 4GL legacy name.
    *           
    * @return   <code>true</code> if the given handle is valid.
    */
   private static boolean validServer(handle hServer, handle hProc, character name)
   {
      final String err2 =
         "Invalid or inappropriate server handle specified for RUN %s ... ON SERVER statement";
      final int code2 = 5453;

      if (hServer.isUnknown())
      {
         ErrorManager.recordOrThrowError(code2, err2, false);
         return false;
      }
      
      Object resource = hServer.get();
      if (!(resource instanceof ServerImpl))
      {
         final String err =
            "The handle for \"RUN ... ON SERVER\" must be either a SERVER handle or " +
            "a SESSION handle";
         ErrorManager.recordOrThrowError(8998, err, false);

         return false;
      }
      
      ServerImpl server = (ServerImpl) resource;
      if (!server._connected())
      {
         final String err1 = "SERVER  is not connected";
         final int code1 = 5451;

         int[] codes = null;
         String[] msgs = null;
         
         if (hProc != null && hProc._isValid() &&
             (ProcedureManager.isProxy(hProc) || hProc.get() instanceof PortTypeWrapper))
         {
            codes = new int[] { code1 };
            msgs = new String[] { err1 };
         }
         else
         {
            codes = new int[] { code1, code2 };
            msgs = new String[] { err1, String.format(err2, name.toStringMessage()) };
         }
         
         ErrorManager.recordOrThrowError(codes, msgs, false);
         
         return false;
      }
      
      return true;
   }
   
   /**
    * Common helper to display a standard error message and then throw a
    * STOP condition when the target procedure cannot be found. Silent
    * error mode has no effect on this since this is not an ERROR condition.
    * That means that this method never returns normally AND that the
    * ERROR-STATUS data is unchanged.
    *
    * @param    target    
    *           The procedure that cannot be found.
    */
   private static void invokeFailure(String target)
   throws StopConditionException
   {
      String spec = "\"%s\" was not found";
      String err  = String.format(spec, target);
      
      ErrorManager.displayError(293, err);
      throw new StopConditionException(err, new NumberedException(err, 293));
   }   
   
   /**
    * Throw a {@link RuntimeException}, as the function or procedure 
    * invocation has raised an exception.
    * 
    * @param    caller
    *           {@link InternalEntryCaller} caller instance used to invoke the
    *           function or procedure. May be null. 
    * @param    name
    *           The legacy 4GL name passed to the function or procedure call
    *           statement.
    * @param    deferred
    *           The deferred error.
    */
   private static void invokeError(InternalEntryCaller caller, 
                                   String name,
                                   Exception deferred)
   {
      if (caller != null)
      {
         throw new RuntimeException("invoke() of class "   +
                                    caller.getClassName()  +
                                    " and  method "        +
                                    caller.getMethodName() +
                                    " failed",
                                    deferred);
      }
      else
      {
         throw new RuntimeException("invoke() of program "   +
                                    name + " failed", deferred);
      }
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Stores the most recent RETURN-VALUE for this user's session. */
      private character returnValue = new character("");

      /** 
       * List of {@link Resolver} instances used for plain RUN statements.
       */
      private final List<Resolver> resolvers;

      /**
       * List of {@link Resolver} instances used for RUN IN handle,
       * DYNAMIC-FUNCTION(... IN handle) or FUNCTION ... IN handle cases.
       */
      private final List<Resolver> inHandleResolvers;
      
      /**
       * {@link InternalResolver} instance which searches the internal-entries
       * for a match.
       */
      private final Resolver internalResolver;
      
      /**
       * {@link InternalProcSuperResolver} instance which searches the 
       * procedure's super-procedures for a match.
       */
      private final Resolver internalProcSuperResolver;
      
      /**
       * {@link SessionSuperResolver} instance which searches the sessions's 
       * super-procedures for a match.
       */
      private final Resolver sessionSuperResolver;
      
      /**
       * {@link ExternalProgramResolver} instance which searches the external
       * program list for a match.
       */
      private final Resolver externalResolver;

      private WorkArea()
      {
         internalResolver = new InternalResolver();
         internalProcSuperResolver = new InternalProcSuperResolver();
         sessionSuperResolver = new SessionSuperResolver();
         externalResolver = new ExternalProgramResolver();
         
         resolvers = new ArrayList<Resolver>();
         resolvers.add(internalResolver);
         resolvers.add(internalProcSuperResolver);
         resolvers.add(sessionSuperResolver);
         resolvers.add(externalResolver);

         inHandleResolvers = new ArrayList<Resolver>();
         inHandleResolvers.add(internalResolver);
         inHandleResolvers.add(internalProcSuperResolver);
         inHandleResolvers.add(sessionSuperResolver);
      }
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }   
   }

   /**
    * Invoke the configured internal entry.
    */
   private static class InternalEntryCaller
   {
      /** 
       * The external program instance where the internal entry's body is
       * defined.
       */
      protected Object callerInstance;
      
      /** The java name of this internal entry. */
      protected String methodName;
      
      /** The legacy 4GL name of this internal entry. */
      protected String iename;
      
      /** The {@link Method} instance for this internal entry. */
      private Method mthd = null;
      
      /** The handle associated with the {@link #callerInstance}. */
      private handle phandle;

      /**
       * Basic c'tor.
       */
      public InternalEntryCaller()
      {
      }
      
      /**
       * Configure this instance to invoke the given internal entry.
       * 
       * @param    phandle
       *           The procedure handle where the internal entry's body is
       *           defined.
       * @param    methodName
       *           The resolved java name for this internal entry.
       * @param    iename
       *           The legacy 4GL name of this internal entry.
       */
      public InternalEntryCaller(handle phandle,
                                 String methodName, 
                                 String iename)
      {
         this.phandle = new handle(phandle);
         this.callerInstance = phandle.get();
         this.methodName = methodName;
         this.iename = iename;
      }
      
      /**
       * Invoke the internal entry, using the {@link #mthd}, 
       * the {@link #callerInstance} and the passed arguments.
       *
       * @param    modes
       *           An encoded string with the mode of each parameter specified as a
       *           character in the string.
       * @param    param
       *           The arguments for this Java method call.
       *           
       * @return   The returned value
       * 
       * @throws   Exception
       *           If any exceptions are encountered during method invocation.
       */
      Object invoke(String modes, Object... param)
      throws Exception
      {
         mthd.setAccessible(true);
         return mthd.invoke(callerInstance, param);
      }

      /**
       * Get the signature, as defined by the parameters.
       * 
       * @param    param
       *           The arguments for this Java method call.
       *           
       * @return   An array with the type of each parameter.
       */
      Class<?>[] getSignature(Object... param)
      {
         Class<?>[] signature = new Class[param.length];
         
         for (int i = 0; i < signature.length; i++)
         {
            Class<?> type = param[i] == null ? Object.class : param[i].getClass();
            if (ProxyFactory.isProxyClass(type) || Proxy.isProxyClass(type))
            {
               // the only provided support is for buffer objects.
               Class<?> dmoBufIfc = RecordBuffer.getDMOBufInterfaceForObject(param[i]);
               
               if (dmoBufIfc == null)
               {
                  throw new RuntimeException(
                     "Dynamic proxy not supported for parameter " + param[i]);
               }
               
               type = dmoBufIfc;
            }
            signature[i] = type;
         }
         
         return signature;
      }
      
      /**
       * Validate the given parameters against the method's real signature.
       * If the parameter's type is compatible with the real type (from the
       * {@link #mthd method's} arguments), then cast it to it. 
       * If all are valid, then {@link ArgValidationErrors#NO_ERROR} is 
       * returned. Else:
       * <ul>
       * <li>if the argument count does not match, then 
       *     {@link ArgValidationErrors#INCORRECT_COUNT} is returned.</li>
       * <li>if the types of the arguments do not match, then 
       *     {@link ArgValidationErrors#INCORRECT_TYPES} is returned.</li>
       * </ul>
       * 
       * @param    function
       *           <code>true</code> if this is a function call.
       * @param    param
       *           The arguments for this Java method call.
       *           
       * @return   One of the {@link ArgValidationErrors} constants.
       */
      ArgValidationErrors valid(boolean function, Object... param)
      {
         Class<?>[] signature = getSignature(param);
         Class<?> cls = callerInstance.getClass();
         
         Method method = null;
         try
         {
            // TODO: cache the methods by class/name/signature....
            method = cls.getMethod(methodName, signature);
            mthd = method;
            return ArgValidationErrors.NO_ERROR;
         }
         catch (Exception e)
         {
            method = null;
         }
         
         // fallback - check if the method is declared, but private
         try
         {
            method = cls.getDeclaredMethod(methodName, signature);
            mthd = method;
            return ArgValidationErrors.NO_ERROR;
         }
         catch (Exception e)
         {
            method = null;
         }
         
         // TODO: add support for the longchar-to-char narrowing and char-to-longchar widening
         
         final Class<?> oep = AbstractExtentParameter.class;
         final String extProgName = ProcedureManager.getAbsoluteName(callerInstance);
         
         // go through all the methods
         boolean correctNo = false;
         boolean hasParams = false;
         for (Method m : cls.getDeclaredMethods())
         {
            if (m.getName().equals(methodName))
            {
               Class<?>[] mtypes = m.getParameterTypes();
               hasParams = (mtypes.length > 0);
               
               if (mtypes.length == param.length)
               {
                  correctNo = true;
                  
                  // try to cast each parameter to the expected type. if
                  // it works, then use this method
                  boolean ok = true;
                  Object[] newPars = new Object[param.length];
                  for (int i = 0; i < param.length; i++)
                  {
                     newPars[i] = param[i];
                     Class<?> pcls = param[i].getClass();

                     Class<?> ptype = signature[i];
                     Class<?> expectedType = mtypes[i];
                     
                     // check if passing a non-array to an array or an array to an array. do not 
                     // check arrays containing non-BDT components
                     if ((ptype.isArray() != expectedType.isArray() && 
                          (!ptype.isArray() || 
                           BaseDataType.class.isAssignableFrom(ptype.getComponentType()))) ||
                         oep.isAssignableFrom(ptype) != oep.isAssignableFrom(expectedType))
                     {
                        boolean parExtent = ptype.isArray() || oep.isAssignableFrom(ptype);
                        
                        int expectedExtent = SourceNameMapper.getExtentForParam(extProgName,
                                                                                iename,
                                                                                function,
                                                                                i);
                        int paramExtent = 0;
                        if (parExtent)
                        {
                           BaseDataType[] pvar = null;
                           if (oep.isAssignableFrom(ptype))
                              pvar = ((AbstractExtentParameter) newPars[i]).getVariable();
                           else
                              pvar = (BaseDataType[]) newPars[i];

                           paramExtent = pvar.length;
                        }
                        
                        ArgValidationErrors err = null;
                        
                        if (expectedExtent == 0)
                        {
                           // expected param is dynamic-extent
                           err = parExtent 
                                    ? ArgValidationErrors.DYN_ARRAY_FOR_NON_ARRAY 
                                    : ArgValidationErrors.NON_ARRAY_FOR_DYN_ARRAY;
                        }
                        else
                        {
                           // expected param is fixed-extent
                           err = parExtent 
                                    ? ArgValidationErrors.ARRAY_FOR_NON_ARRAY
                                    : ArgValidationErrors.NON_ARRAY_FOR_ARRAY;
                        }
                        
                        err.expectedExtent = parExtent ? 0 : expectedExtent;
                        err.paramExtent = parExtent ? paramExtent : 0;
                        
                        // TODO: finish all cases of extent parameter validation
                        
                        return err;
                     }
                     
                     // provide compatibility between resources and handles
                     if (WrappedResource.class.isAssignableFrom(ptype) && 
                         handle.class.equals(expectedType))
                     {
                        newPars[i] = new handle((WrappedResource) param[i]);
                        continue;
                     }
                     
                     if (!expectedType.isAssignableFrom(ptype))
                     {
                        try
                        {
                           // if the param's type is unknown class, construct
                           // the unknown for the expectedType
                           if (unknown.class.equals(ptype))
                           {
                              // for some data types (date, datetime, datetime-tz, raw) the 
                              // unknown value is not created using default constructor
                              newPars[i] = BaseDataType.generateUnknown(expectedType);
                           }
                           else
                           {
                              Constructor<?> ctor = expectedType.getConstructor(pcls);
                              newPars[i] = ctor.newInstance(param[i]);
                           }
                           if (ErrorManager.isPending())
                           {
                              // if any errors occurred during parameter matching 
                              // (eg.: INTEGER overflow) procedure cannot be called
                              ok = false;
                              break;
                           }
                        }
                        catch (Throwable t)
                        {
                           // was not able to convert this param
                           ok = false;
                           
                           // check if a condition was raised. if so, do not
                           // search for any other c'tors
                           boolean raisedCondition = false;
                           while (t != null)
                           {
                              if (t instanceof ConditionException)
                              {
                                 raisedCondition = true;
                              }
                              t = t.getCause();
                           }
                           
                           if (raisedCondition)
                           {
                              break;
                           }
                        }

                        if (!ok)
                        {
                           // fallback, maybe there is a c'tor with a
                           // compatible type
                           Constructor<?>[] ctors = expectedType.getConstructors();
                           for (Constructor<?> ctor : ctors)
                           {
                              Class<?>[] ctypes = ctor.getParameterTypes();
                              if (ctypes.length == 1 && ctypes[0].isAssignableFrom(pcls))
                              {
                                 // found my c'tor
                                 try
                                 {
                                    newPars[i] = ctor.newInstance(param[i]);
                                    ok = true;
                                    break;
                                 }
                                 catch (Exception e)
                                 {
                                    // ignore
                                 }
                              }
                           }
                        }

                        // if still wasn't found, then exit.
                        if (!ok)
                        {
                           break;
                        }
                     }
                  }

                  if (ok)
                  {
                     // was able to cast the parameters to expected types
                     for (int i = 0; i < param.length; i++)
                     {
                        param[i] = newPars[i];
                     }
                     mthd = m;

                     return ArgValidationErrors.NO_ERROR;
                  }
               }
            }
         }
         
         return !correctNo ? (!hasParams ? ArgValidationErrors.UNEXPECTED_PARAM 
                                         : ArgValidationErrors.INCORRECT_COUNT)
                           : ArgValidationErrors.INCORRECT_TYPES;
      }
      
      /**
       * Get the java method name for this internal entry.
       * 
       * @return   See above.
       */
      String getMethodName()
      {
         return methodName;
      }
      
      /**
       * Get the class name of the external program to which this internal
       * entry belongs.
       * 
       * @return   See above.
       */
      String getClassName()
      {
         return callerInstance.getClass().getName();
      }
      
      /**
       * Get the external program instance where this internal entry is 
       * defined.
       * 
       * @return   See above.
       */
      Object getCallerInstance()
      {
         return callerInstance;
      }
      
      /**
       * Get the legacy name of this internal entry.
       * 
       * @return   See above.
       */
      String getInternalEntryName()
      {
         return iename;
      }
      
      /**
       * Get the string representation of the parameter modes defined for this
       * internal entry.
       * 
       * @param    function
       *           <code>true</code> if this is a function call.
       * 
       * @return   See above.
       */
      String getParameterModes(boolean function)
      {
         String pname = ProcedureManager.getAbsoluteName(callerInstance);
         String iename = getInternalEntryName();
         return SourceNameMapper.getParameterModes(pname, iename, function);
      }
      
      /**
       * Get the {@link #phandle} field.
       * 
       * @return   See above.
       */
      handle getPhandle()
      {
         return phandle;
      }
   }

   /**
    * This caller is returned when the FUNCTION ... MAP TO can not be resolved. It contains the 
    * function name from the last MAP TO clause.
    */
   private static class MapToNotFound
   extends InternalEntryCaller
   {
      public MapToNotFound(String iename)
      {
         this.iename = iename;
      }
   }
   
   /**
    * Implements native procedure invocation (an API call to a shared library).
    */
   private static class NativeProcedureCaller
   extends InternalEntryCaller
   {
      /** The library name where this native procedure belongs. */
      private String libname = null;
      
      /** The legacy name of this native procedure. */
      private String pname = null;
      
      /**
       * Configure this caller with the given data.
       * 
       * @param    libname
       *           The library name where this native procedure should be found.
       * @param    pname
       *           The legacy name of the external procedure from which this API is being called.
       * @param    iename
       *           The internal-entry name being called which is usually the same as the entry
       *           point name in the library except where an ordinal is being used.
       */
      public NativeProcedureCaller(String libname, String pname, String iename)
      {
         this.iename  = iename;
         this.libname = libname;
         this.pname   = pname;
      }
      
      /**
       * Call the native API with the arguments provided.
       *
       * @param    modes
       *           An encoded string with the mode of each parameter specified as a
       *           character in the string.
       * @param    args
       *           The array of procedure parameters.
       *
       * @return   Always <code>null</code> since procedures don't have a return value in the
       *           4GL.
       */
      @Override
      Object invoke(String modes, Object... args) 
      throws Exception
      {
         Class<?>[] signature = getSignature(args);
         NativeInvoker.invoke(this.libname, this.pname, this.iename, signature, modes, args);
         return null;
      }
      
      /**
       * For native API calls, the validation is deferred to the actual time of the call since
       * the 4GL does other parameter checking before the more traditional type and mode
       * checking.  Rather than place that logic here, the decision was made to keep all of
       * that processing together in {@link NativeInvoker#invoke}.
       * 
       * @param    function
       *           <code>true</code> if this is a function call.
       * @param    param
       *           The arguments for this Java method call.
       *           
       * @return   Always returns {@link ArgValidationErrors#DEFER_VALIDATION}, which disables
       *           validation processing at this time.
       */
      ArgValidationErrors valid(boolean function, Object... param)
      {
         return ArgValidationErrors.DEFER_VALIDATION;
      }
      
      /**
       * Get the java method name for this internal entry.
       * 
       * @return   Always the same as the legacy native function name.  This is OK since
       *           this method is only used for error messages.
       */
      String getMethodName()
      {
         return pname;
      }
      
      /**
       * Get the class name of the external program to which this internal entry belongs.
       * 
       * @return   Always a nonsense class name.   This is OK since this method is only used
       *           for error messages.
       */
      String getClassName()
      {
         return "NATIVE_LIBRARY_CALL";
      }
   }

   /**
    * Implements external program invocation.
    */
   private static class ExternalProcedureCaller
   extends InternalEntryCaller
   {
      /** The class name associated with the external program. */
      private String className;

      /**
       * Associate this caller with the specified external program instance.
       * 
       * @param    instance
       *           An instance of an external program.
       */
      public ExternalProcedureCaller(Object instance)
      {
         this.methodName = "execute";
         this.className = instance.getClass().getName();

         this.callerInstance = instance;
      }
      
      
      /**
       * Configure this caller with the given data.
       * 
       * @param    className
       *           The class name associated with the external program.
       *           
       * @throws   Exception
       *           If the class can't be resolved. 
       */
      public ExternalProcedureCaller(String className)
      throws Exception
      {
         this.methodName = "execute";
         this.className = className;

         Class<?> cls = Class.forName(className);
         this.callerInstance = cls.newInstance();
      }

      @Override
      String getClassName()
      {
         return className;
      }

      @Override
      String getInternalEntryName()
      {
         return "";
      }
      
      @Override
      String getParameterModes(boolean function)
      {
         String pname = ProcedureManager.getAbsoluteName(callerInstance);
         return SourceNameMapper.getParameterModes(pname);
      }
      
      @Override
      handle getPhandle()
      {
         // For external procedures, the phandle field is always null.
         return null;
      }
   }

   /**
    * Abstract class. Provides APIs to determine the correct target of a 
    * function or procedure invocation statement. 
    */
   private static abstract class Resolver
   {
      /**
       * Based on the given data, determine the procedure handle to which the
       * internal entry belongs and its java method name (or the external
       * program to be invoked).
       * 
       * @param    phandle
       *           The procedure handle where this internal entry is defined.
       * @param    name
       *           The legacy name of the internal entry or external program.
       * @param    function
       *           <code>true</code> if need to search for functions. 
       *           <code>false</code> if need to search for procedures.
       * @param    superCall
       *           <code>true</code> if this is a RUN SUPER or SUPER() call.
       * @param    exports
       *           The list of allowed external procedures to be invoked with a RUN ... ON SERVER
       *           statement. 
       *           
       * @return   An {@link InternalEntryCaller caller} or null if not found.
       */
      abstract InternalEntryCaller resolve(handle    phandle, 
                                           String    name,
                                           boolean   function,
                                           boolean   superCall,
                                           character exports);
   }

   /** 
    * Search through the handle's defined internal-entries for a match.
    */
   private static class InternalResolver
   extends Resolver
   {
      @Override
      public InternalEntryCaller resolve(handle    phandle, 
                                         String    ieName, 
                                         boolean   function,
                                         boolean   superCall,
                                         character exports)
      {
         String pname = ProcedureManager.getAbsoluteName(phandle.get());

         String methodName = 
            SourceNameMapper.getMethodName(pname, ieName, function);
         if (methodName == null)
         {
            return null;
         }

         if (SourceNameMapper.isPrivate(pname, ieName, function))
         {
            handle thisp = ProcedureManager.thisProcedure();
            if (!CompareOps.equals(thisp, phandle))
            {
               return null;
            }
         }

         InternalEntryCaller caller = null;
         
         String libname = SourceNameMapper.getLibraryName(pname, ieName);
         String mapTo   = (function ? SourceNameMapper.getMapTo(pname, ieName) : null);
         
         if (libname != null)
         {
            caller = new NativeProcedureCaller(libname, pname, ieName);
         }
         else if (mapTo != null)
         {
            // go recursively
            InternalEntryCaller res = resolve(phandle, mapTo, function, superCall, exports);
            
            if (res == null)
            {
               res = new MapToNotFound(mapTo);
            }
            
            return res;
         }
         else if (!SourceNameMapper.isSuper(pname, ieName, function))
         {
            handle hcaller = phandle;
            
            if (SourceNameMapper.isInHandle(pname, ieName, function))
            {
               hcaller = ProcedureManager.getHandleForFunction(phandle,
                                                               ieName);
            }
            
            caller = new InternalEntryCaller(hcaller, methodName, ieName);
         }
         /*
         else if (!superCall)
         {
            // IN SUPER functions and procedures are used only when this
            // call is not from a RUN SUPER or SUPER call.
            Resolver r = work.obtain().internalProcSuperResolver;
            return r.resolve(phandle, ieName, function, superCall);
         }*/
         
         return caller;
      }
   }

   /**
    * Search through the given super-procedure list for a match.
    */
   private static abstract class SuperResolver
   extends InternalResolver
   {
      private final Set<Object> used = new HashSet<Object>();
      
      protected InternalEntryCaller resolveSuperProc(List<handle> superProcs,
                                                     String       ieName,
                                                     boolean      function,
                                                     boolean      superCall,
                                                     character    exports)
      {
         // search is done backwards (last set super-procedure is searched
         // first)
         for (handle sphandle : superProcs)
         {
            Object referent = sphandle.get();
            
            if (used.contains(referent))
            {
               continue;
            }
            
            used.add(referent);
            try
            {
               // when looking for an internal-entry, only the direct super-
               // procedures are searched. no recursivity is done.
               InternalEntryCaller caller = super.resolve(sphandle,
                                                          ieName,
                                                          function,
                                                          superCall,
                                                          exports);
               if (caller != null)
               {
                  return caller;
               }
            }
            finally
            {
               used.remove(referent);
            }
         }
         
         return null;
      }
   }

   /**
    * Search through the super-procedures of the given procedure handle for
    * a match.
    */
   private static class InternalProcSuperResolver
   extends SuperResolver
   {
      @Override
      public InternalEntryCaller resolve(handle    phandle,
                                         String    ieName,
                                         boolean   function,
                                         boolean   superCall,
                                         character exports)
      {
         List<handle> superProcs = ProcedureManager.getSuperProcedures(phandle.get());

         return resolveSuperProc(superProcs, ieName, function, superCall, exports);
      }
   }

   /**
    * Search through the session's super-procedures of the handle for a match.
    */
   private static class SessionSuperResolver
   extends SuperResolver
   {
      @Override
      public InternalEntryCaller resolve(handle    phandle,
                                         String    ieName,
                                         boolean   function,
                                         boolean   superCall,
                                         character exports)
      {
         List<handle> superProcs = ProcedureManager.getSuperProcedures();

         return resolveSuperProc(superProcs, ieName, function, superCall, exports);
      }
   }

   /**
    * Search through the available external programs for a match.
    */
   private static class ExternalProgramResolver
   extends Resolver
   {
      /** Logger */
      private static final Logger LOG = LogHelper.getLogger(ExternalProgramResolver.class);
      
      @Override
      public InternalEntryCaller resolve(handle    phandle,
                                         String    name,
                                         boolean   function,
                                         boolean   superCall,
                                         character exports)
      {
         if (function)
         {
            // for functions, don't search external procedures
            return null;
         }
         
         if (exports != null      &&
             !exports.isUnknown() &&
             !TextOps.matchesList(exports, name).booleanValue())
         {
            // in case this is call is via an appserver, only the exported procedures can be used
            return null;
         }
         
         String className = SourceNameMapper.convertName(name);
         
         InternalEntryCaller caller = null;
         if (className != null)
         {
            // is a registered external procedure
            try
            {
               caller = new ExternalProcedureCaller(className);
            }
            catch (Throwable t)
            {
               LOG.log(Level.WARNING, "Unable to resolve external program.", t);
               
               Throwable cause = t;
               while (cause != null)
               {
                  // if the exception is a condition exception, let it 
                  // propagate up the stack
                  if (cause instanceof ConditionException)
                  {
                     throw (ConditionException) cause;
                  }
                  if (cause instanceof RetryUnwindException)
                  {
                     throw (RetryUnwindException) cause;
                  }

                  cause = cause.getCause();
               }

               return null;
            }
         }

         return caller;
      }
   }

   /**
    * Helper class which also saves the {@link AsyncRequestImpl} instance, needed by the remote
    * request.
    */
   private static abstract class AsyncCall
   implements Runnable
   {
      /** The async request instance for this call. */
      public final AsyncRequestImpl asyncReq;

      /** The server used to execute this async call. */
      public final handle hServer;
      
      /** A handle variable where to save the web service. May be null. */
      public final handle portHandle;
      
      /** The target procedure name. */
      public final character pName;
      
      /** The handle to which the procedure belongs, if it is RUN ... IN handle statement. */
      public final handle pHandle;
      
      /** The handle to the asynchronous request. */
      public final handle asyncHandle;
      
      /** This represents the name of the internal procedure for the EVENT-PROCEDURE clause. */
      public final character eventProcName;
      
      /** The handle to a procedure in which context the EVENT-PROCEDURE is found. */
      public final handle inEventprocHandle;
      
      /**
       * This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON statement.
       */
      public final boolean transactionDistinct;
      
      /**
       * 
       * @param    hServer
       *           The server used to execute this async call.
       * @param    portHandle
       *           A handle variable where to save the web service. May be null.
       * @param    pName
       *           The target procedure name.
       * @param    pHandle
       *           The handle to which the procedure belongs, if it is RUN ... IN handle 
       *           statement.
       * @param    asyncHandle
       *           The handle to the asynchronous request.
       * @param    eventProcName
       *           This represents the name of the internal procedure for the EVENT-PROCEDURE 
       *           clause, null if this clause is not specified.
       * @param    inEventprocHandle
       *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
       *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
       * @param    transactionDistinct
       *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON 
       *           statement.
       * @param    modes
       *           A string representation of the modes of each parameter.
       * @param    args
       *           The procedure's arguments.
      */
      public AsyncCall(handle     hServer,
                       handle     portHandle,
                       character  pName,
                       handle     pHandle,
                       handle     asyncHandle,
                       character  eventProcName,
                       handle     inEventprocHandle,
                       boolean    transactionDistinct,
                       String     modes,
                       Object...  args)
      {
         this.hServer = new handle(hServer);
         this.portHandle = portHandle == null ? null : new handle(portHandle);
         this.pName = new character(pName);
         this.pHandle = pHandle == null ? null : new handle(pHandle);
         this.asyncHandle = asyncHandle == null ? null : new handle(asyncHandle);
         this.eventProcName = new character(eventProcName);
         this.inEventprocHandle = new handle(inEventprocHandle);
         this.transactionDistinct = transactionDistinct;

         this.asyncReq = new AsyncRequestImpl(this.hServer,
                                              this,
                                              this.eventProcName,
                                              this.inEventprocHandle,
                                              this.pName,
                                              this.pHandle,
                                              modes,
                                              args);
      }
   }
}
